"""Implements a variety of prediction models."""
from collections import Counter
from copy import deepcopy
from typing import Union


def principal_delta(principal, other, potential):
    modification = abs(principal - other) * potential
    if other < principal:
        return principal - modification
    elif other > principal:
        return principal + (abs(other - principal) * potential)


def model_per_emotive(ensemble: list, emotive: str, potential_normalization_factor: float) -> float:
    """Compute the modeled emotive value, using the prediction ensemble and potential_normalization_factor

    Identifies the principal value for the emotive by utilizing the first prediction (e.g. highest potential prediction)
    containing that emotive. For the rest of predictions in the ensemble, the modeled value is updated using the delta
    between the current emotive value and the principal value * current potential / potential_normalization_factor.
    
    The intent is to ensure that the resultant modeled emotive value is never greater than the individual value of an emotive,
    and that subsequent emotive values impact the modeled value less and less, based on their potential values.
    
    Prior to updating this function, the delta to update the modeled value was computed by subtracting the current emotive value
    from the principal value, but this resulted in a case where the modeled emotive value was greater than any individual value seen
    in the prediction ensemble.
    
    Args:
        ensemble (list): The prediction ensemble to use in computations
        emotive (str): The emotive to compute a moving average of
        potential_normalization_factor (float): Sum of potential from all predictions in ensemble

    Returns:
        float: modeled emotive value for specific emotive
    """
    # using a weighted posterior_probability = potential/marginal_probability
    # FORMULA: pv + ( (Uprediction_2-pv)*(Wprediction_2) + (Uprediction_3-pv)*(Wprediction_3)... )/mp
    _found = False
    principal_value = 0.0
    while not _found:
        for i in range(0, len(ensemble)):
            if emotive in ensemble[i]['emotives'].keys():
                _found = True
                principal_value = ensemble[i]['emotives'][emotive]  # Let's use the "best" match (i.e. first showing of this emotive) as our starting point. Alternatively, we can use,say, the average of all values before adjusting.
                break
        if i == len(ensemble):
            return principal_value
    v = principal_value
    for x in ensemble[i + 1:]:
        if emotive in x['emotives']:
            v += (x["potential"] / potential_normalization_factor) * (x['emotives'][emotive] - v)
    return v


def average_emotives(record: list) -> dict:
    """
    Averages the emotives in a list (e.g. predictions ensemble or percepts).
    The emotives in the record are of type: [{'e1': 4, 'e2': 5}, {'e2': 6}, {'e1': 5 'e3': -4}]

    Args:
        record (list): List of emotive dictionaries to average emotives of

    Returns:
        dict: Dictionary of Averaged Emotives

    Example:
        ..  code-block:: python

            from ia.gaius.prediction_models import average_emotives
            record = [{'e1': 4, 'e2': 5}, {'e2': 6}, {'e1': 5 'e3': -4}]
            averages = average_emotives(record=record)

    """
    new_dict = {}
    for bunch in record:
        for e, v in bunch.items():
            if e not in new_dict:
                new_dict[e] = [v]
            else:
                new_dict[e].append(v)
    avg_dict = {}
    for e, v in new_dict.items():
        avg_dict[e] = float(sum(v) / len(v))
    return avg_dict


def bucket_predictions(ensemble: list) -> list:
    """Buckets predictions with identical potential values
    into a single prediction, merging emotives dicts and averaging
    as necessary.
    
    This function is invoked before :func:`model_per_emotive` to ensure that there is only
    a single prediction corresponding to each potential level. This is because the model_per_emotive
    function computes a delta between predictions based on potential value. By flattening to a single
    prediction, we ensure that there are no collsions in the principal_value computed in model_per_emotive
    based on potential.

    Args:
        ensemble (list): Prediction ensemble

    Returns:
        list: Resultant ensemble
    """
    bucket_dict = {}

    for pred in ensemble:

        if pred['potential'] in bucket_dict.keys():
            bucket_dict[pred['potential']].append(pred)
        else:
            bucket_dict[pred['potential']] = [pred]

    new_ensemble = []
    for v in bucket_dict.values():

        singular_pred = v[0]
        singular_pred['emotives'] = average_emotives([p['emotives'] for p in v])

        new_ensemble.append(singular_pred)

    return new_ensemble


def prediction_ensemble_modeled_emotives(ensemble: list) -> dict:
    """The emotives in the ensemble are of type: 'emotives':[{'e1': 4, 'e2': 5}, {'e2': 6}, {'e1': 5 'e3': -4}]
    First calls :func:`average_emotives` on each prediction in the ensemble, then calls :func:`bucket_predictions` on the ensemble.
    After bucketing predictions, the function :func:`model_per_emotives` is called for each emotive present in the ensemble.
    Dict returned contains { emotive: :func:`model_per_emotive` } for each emotive in the ensemble

    The potential_normalization_factor is taken after bucketing predictions to ensure that all predictions with identical potential values
    are flattened into a single list (avoids collisions with principal_values in :func:`model_per_emotive`)
    Args:
        ensemble (list): Prediction ensemble containing emotives to model

    Returns:
        dict: Dictionary of modelled emotive values
    """

    emotives_set = set()

    filtered_ensemble = []
    for p in ensemble:
        new_record = deepcopy(p)
        new_record['emotives'] = average_emotives([new_record['emotives']])
        filtered_ensemble.append(new_record)

    filtered_ensemble = bucket_predictions(filtered_ensemble)

    potential_normalization_factor = sum([p['potential'] for p in filtered_ensemble])

    for p in filtered_ensemble:
        emotives_set = emotives_set.union(p['emotives'].keys())
    return {emotive: model_per_emotive(ensemble, emotive, potential_normalization_factor) for emotive in emotives_set}


def hive_model_emotives(ensembles: dict) -> dict:
    """Compute average of emotives in model by calling :func:`average_emotives` on `ensembles`
    Internally calls :func:`prediction_ensemble_modeled_emotives` on each ensemble,
    then calls :func:`average_emotives` on the results
    Args:
        ensembles (dict): should be dictionary of { node_name: prediction_ensemble }

    Returns:
        dict: Dictionary of Averaged Emotives
    """
    return average_emotives([prediction_ensemble_modeled_emotives(ensemble) for ensemble in ensembles.values()])


def prediction_ensemble_model_classification(ensemble: list, strip_pipes: bool = True) -> Union[Counter, None]:
    """Compute classification rankings based on the symbols present in a prediction ensemble

    Args:
        ensemble (list): Prediction Ensemble
        strip_pipes (bool): Strip pipes from classifications

    Returns:
        Counter: dictionary containing ranking of classification symbols
    """
    boosted_prediction_classes = Counter()
    for prediction in ensemble:
        if not prediction['future']:
            continue
        symbol: str
        for symbol in prediction['future'][-1]:
            if strip_pipes and "|" in symbol:
                # grab the value, remove the piped keys
                symbol = symbol.split("|")[-1]
            boosted_prediction_classes[symbol] += prediction['potential']
    if len(boosted_prediction_classes) > 0:
        return boosted_prediction_classes
    else:
        return None


def most_common_ensemble_model_classification(ensemble: list, strip_pipes: bool = True) -> Union[str, None]:
    result = prediction_ensemble_model_classification(
        ensemble=ensemble, strip_pipes=strip_pipes)
    if result is None:
        return result
    return result.most_common(1)[0][0]


def hive_model_classification(ensembles: dict, strip_pipes: bool = True) -> Union[Counter, None]:
    """Compute the "hive predicted model classification" based on the ensembles provided from each node

    Args:
        ensembles (dict): should be dictionary of { node_name: prediction_ensemble }
        strip_pipes (bool): Strip pipes from classifications

    Returns:
        str: hive predicted classification
    """
    if ensembles:
        # This just takes the first "most common", even if there are multiple that have the same frequency.
        boosted_classifications = [prediction_ensemble_model_classification(
            c, strip_pipes=strip_pipes) for c in ensembles.values()]
        votes = Counter()
        for p in boosted_classifications:
            votes.update(p)
        if votes:
            return votes
    return None



from typing import Dict, List, Any, Tuple, Union, Callable
import requests
from requests import Response
import logging

logger = logging.getLogger(__name__)


class TFQueryError(BaseException):
    """Raised if any query to TF returns an error."""
    pass


# example tf_info
# { "name": "MainTF",
#   "url": "https://tf.intelligent-artifacts.com:8090",
#   "api_key": "ABCD-1234"
# }

class TFClient:
    """Thinkflux Client Wrapper (interface to TF REST API)
    """

    def __init__(self, tf_info: dict, verify: bool = True) -> None:

        self._tf_info = tf_info

        self._url = tf_info['url']
        self._api_key = tf_info['api_key']
        self.name = tf_info['name']

        self.session = requests.Session()
        self.session.headers['X-API-KEY'] = self._api_key
        self.verify = verify
        self.session.verify = verify

    def set_verify(self, verify: bool):
        self.verify = verify
        self.session.verify = verify

    def _query(self,
               query_method: Callable,
               path: str,
               timeout: float = None,
               **kwargs) -> Response:
        response = query_method(path, verify=self.verify,
                                timeout=timeout, **kwargs)
        if response.status_code != 200:
            raise TFQueryError(
                f"Bad response {response} from TF: {response.content}")
        return response

    def ping(self) -> str:
        response = requests.get(self._url + '/ping', verify=self.verify)
        if response.status_code != 200:
            raise TFQueryError("Failed to ping Thinkflux")
        return response.text

    def show_status(self) -> dict:

        return self._query(requests.get, self._url + '/show_status').json()

    def clear_all_kbs(self) -> dict:

        return self._query(requests.get, self._url + '/clear_all_kbs').json()

    def clear_wm(self) -> dict:

        return self._query(requests.get, self._url + '/clear_wm').json()

    def clear_concepts_and_instances(self) -> dict:

        return self._query(requests.get, self._url + '/clear_concepts_and_instances').json()

    def add_schema_symbol_information(self) -> Response:

        return self._query(requests.get, self._url + '/add_schema_symbol_information')

    def observe(self, data: dict) -> dict:
        """Observe data on Thinkflux

        Args:
            data (dict): data percept to observe

        """

        return self._query(requests.post, self._url + '/observe', json=data).json()

    def add_interface_nodes(self, agent_config: dict) -> dict:
        """Add an interface node from GAIuS agent to Thinkflux

        Args:
            agent_config (dict): dictionary with configuration details for interface node

        agent_config dict needs the following fields:
        {
            "name": str,
            "domain: str,
            "secure": bool,
            "api_key": str,
            "interface_nodes": List[str]
        }

        """

        return self._query(requests.post, self._url + '/add_interface_nodes', json=agent_config).json()

    def list_interface_nodes(self) -> dict:

        return self._query(requests.get, self._url + '/list_interface_nodes').json()

    def delete_interface_nodes(self, data: dict) -> Response:

        return self._query(requests.post, self._url + '/delete_interface_nodes', json=data).json()

    def get_model_association_networks(self) -> Response:

        return self._query(requests.get, self._url + '/get_model_association_networks')

    def timer(self) -> Response:

        return self._query(requests.get, self._url + '/timer')

    def update_schema(self, data: dict) -> Response:

        return self._query(requests.post, self._url + '/update_schema', json=data)

    def bootstrap_concepts(self,
                           hierarchy: bool = False,
                           labelled: bool = False,
                           use_labels_for_concepts: bool = False,
                           use_labels_for_hierarchy: bool = False,
                           subsequence_clustering: bool = False) -> Response:
        json_obj = {'hierarchy': hierarchy,
                    'labelled': labelled,
                    'use_labels_for_concepts': use_labels_for_concepts,
                    'use_labels_for_hierarchy': use_labels_for_hierarchy,
                    'subsequence_clustering': subsequence_clustering}
        return self._query(requests.post, self._url + '/bootstrap_concepts', json=json_obj)

    def investigate(self, data: dict) -> Response:

        return self._query(requests.post, self._url + '/investigate_tf', json=data)

    def dreamer(self, data: dict) -> Response:

        return self._query(requests.post, self._url + '/dreamer', json=data)

    def load_schema_base(self, schema_base: dict) -> Response:

        return self._query(requests.post, self._url + '/load_sb', json=schema_base)

    def clear_all_emotives(self) -> dict:

        return self._query(requests.get, self._url + '/clear_all_emotives').json()

    def get_rules_kb(self) -> dict:

        return self._query(requests.get, self._url + '/get_rules_kb').json()

    def set_rules_kb(self, rules_kb: dict) -> dict:

        return self._query(requests.post, self._url + '/set_rules_kb', json=rules_kb).json()

    def set_symbolic_goal(self, symbolic_goal: dict) -> dict:

        return self._query(requests.post, self._url + '/set_symbolic_goal', json=symbolic_goal).json()

    def get_symbolic_goal(self) -> dict:

        return self._query(requests.get, self._url + '/get_symbolic_goal').json()

    def evaluate_world(self, data: dict) -> dict:

        return self._query(requests.post, self._url + '/evaluate_world', json=data).json()

    def get_plan(self) -> dict:

        return self._query(requests.get, self._url + '/get_plan').json()

    def get_concept(self, concept_name: str) -> dict:

        return self._query(requests.get, self._url + f"/concept/{concept_name}").json()

    def get_concepts(self) -> dict:

        return self._query(requests.get, self._url + f"/get_concepts").json()

    def get_instances(self) -> dict:

        return self._query(requests.get, self._url + f"/get_instances").json()



"""Provides utilities for higher level operations on KnowledgeBases"""
from ia.gaius.agent_client import AgentClient
from collections import Counter, defaultdict
from typing import Dict

def list_models(agent: AgentClient, nodes=None):
    """Return a dict of {node_name: model_list} found on specified nodes

    Args:
        agent (AgentClient): GAIuS Agent
        nodes (list, optional): nodes to list models on

    Returns:
        dict: {node_name: model_list} for each node specified in nodes

    Example:
        .. code-block:: python

            from ia.gaius.agent_client import AgentClient
            from ia.gaius.kb_ops import list_models

            agent = AgentClient(agent_info)

            #get list of models found on node P1
            models = list_models(agent, nodes=['P1'])

    """
    if not agent._connected:
        agent.connect()

    prev_summarize_state = agent.summarize_for_single_node
    try:
        agent.set_summarize_for_single_node(False)
        kb = agent.get_kbs_as_json(nodes=nodes, ids=False, obj=True)
        models_dict = {k: list(v['models_kb'].keys()) for k, v in kb.items()}

    except Exception as e:
        print(f'Error in list_models: {e}')
        raise e
    finally:
        agent.set_summarize_for_single_node(prev_summarize_state)

    return models_dict


def list_symbols(agent: AgentClient, nodes=None):
    """Return a dict of {node_name: symbol_list} found on specified nodes

    Args:
        agent (AgentClient): GAIuS Agent
        nodes (list, optional): nodes to list symbols on

    Returns:
        dict: {node_name: symbol_list} for each node specified in nodes

    Example:
        .. code-block:: python

            from ia.gaius.agent_client import AgentClient
            from ia.gaius.kb_ops import list_symbols

            agent = AgentClient(agent_info)

            #get list of symbols found on node P1
            symbols = list_symbols(agent, nodes=['P1'])

    """
    if not agent._connected:
        agent.connect()

    prev_summarize_state = agent.summarize_for_single_node
    try:
        agent.set_summarize_for_single_node(False)
        kb = agent.get_kbs_as_json(nodes=nodes, ids=False, obj=True)
        symbols_dict = {k: list(v['symbols_kb'].keys()) for k, v in kb.items()}
    except Exception as e:
        print(f'Error in list_symbols: {e}')
        raise e
    finally:
        agent.set_summarize_for_single_node(prev_summarize_state)

    return symbols_dict


def get_models_containing_symbol(agent: AgentClient, symbol_set: set, nodes=None):
    """Checks for presence of symbols from `symbol_set` in each model on `nodes`, adding to return dict if a symbol is found.

    Args:
        agent (AgentClient): GAIuS Agent
        symbol (str): the symbol to search for
        nodes (_type_, optional): nodes to search. Defaults to searching all nodes

    Example:
        .. code-block:: python

            from ia.gaius.agent_client import AgentClient
            from ia.gaius.kb_ops import get_models_containing_symbol

            symbols = ["world"]
            agent = AgentClient(agent_info)

            output = get_models_containing_symbol(agent=agent, symbol_set=set(symbols))
    """

    if not agent._connected:
        agent.connect()

    sym_dict = {}
    prev_summarize_state = agent.summarize_for_single_node
    try:
        agent.set_summarize_for_single_node(False)
        kb = agent.get_kbs_as_json(nodes=nodes, ids=False, obj=True)

        for node, node_kb in kb.items():
            sym_dict[node] = set()
            if 'models_kb' not in node_kb:
                continue
            for model_name, model in node_kb['models_kb'].items():
                for event in model['sequence']:
                    for sym in event:
                        if sym in symbol_set:
                            sym_dict[node].add(model_name)

    except Exception as e:
        print(f'Error in identify_models_with_symbol: {e}')
        raise e
    finally:
        agent.set_summarize_for_single_node(prev_summarize_state)

    return sym_dict


def get_models_containing_symbol_strict(agent: AgentClient, symbol_set: set, nodes=None):
    """Checks for presence of symbols from `symbol_set` in each model on `nodes`.
    Only adds model to return dict if all symbols in model are from symbol set
    Store as a dict of {node_name : list}

    Args:
        agent (AgentClient): GAIuS Agent
        symbol (str): the symbol to search for
        nodes (_type_, optional): nodes to search. Defaults to searching all nodes


    Example:
        .. code-block:: python

            from ia.gaius.agent_client import AgentClient
            from ia.gaius.kb_ops import get_models_containing_symbol_strict

            symbols = ["hello", "world", "goodbye", "cruel"]
            agent = AgentClient(agent_info)

            output = get_models_containing_symbol_strict(agent=agent, symbol_set=set(symbols))

    """

    if not agent._connected:
        agent.connect()

    sym_dict = {}
    prev_summarize_state = agent.summarize_for_single_node
    try:
        agent.set_summarize_for_single_node(False)
        kb = agent.get_kbs_as_json(nodes=nodes, ids=False, obj=True)

        for node, node_kb in kb.items():
            sym_dict[node] = set()
            if 'models_kb' not in node_kb:
                continue
            for model_name, model in node_kb['models_kb'].items():
                unique_symbols_in_model = set()
                for event in model['sequence']:
                    for sym in event:
                        unique_symbols_in_model.add(sym)

                if all([sym in symbol_set for sym in unique_symbols_in_model]):
                    sym_dict[node].add(model_name)

    except Exception as e:
        print(f'Error in identify_models_with_symbol: {e}')
        raise e
    finally:
        agent.set_summarize_for_single_node(prev_summarize_state)

    return sym_dict


def get_kb_subset(agent: AgentClient, model_dict: dict):
    """Retrieve a subset of a Knowledgebase based on the provided model_dict.
    Will only provide used symbols and vectors, all others will be trimmed.

    Args:
        agent (AgentClient): GAIuS Agent
        model_dict (dict): {node_name: model_list}. Expected format is similar to that returned from :func:`list_models`

    Raises:
        e: Exception in subset iteration (e.g. model not found on node, get_kb failed, etc.)

    Returns:
        dict: Subset of Knowledgebases corresponding to provided model_dict

    Example:
        .. code-block:: python
            :force:

            from ia.gaius.kb_ops import get_kb_subset, list_models
            from ia.gaius.agent_client import AgentClient

            agent = AgentClient(agent_info)
            agent.connect()

            models = list_models(agent=agent, nodes=['P1'])

            # get a subset of available models
            models = {k: v[:20] for k,v in models.items()}

            # get a subset of the entire Knowledgebase
            kb_subset = get_kb_subset(agent=agent, model_dict=models)

    """

    try:
        reconstructed_kb = {}
        for node, models in model_dict.items():

            node_kb = agent.get_kbs_as_json(obj=True, nodes=[node], ids=False)
            node_kb = node_kb[node]

            if not all([model in node_kb["models_kb"] for model in models]):
                print(f'Not all models were found on node {node}')
                for model in models:
                    if model not in node_kb["models_kb"]:
                        print(f'model "{model}" not found')
                raise Exception(f'Not all models were found on node {node}')

            reconstructed_kb[node] = {'models_kb': {},
                                      'symbols_kb': {},
                                      'vectors_kb': {},
                                      'metadata': {}}
            print(f'original model count: {len(node_kb["models_kb"])}')
            print(f' reduced model count: {len(models)}')
            # will need to recompute symbol frequency and model member frequency counts
            symbol_freq_counter = Counter()
            symbol_model_member_freq_counter = Counter()

            # only keep symbols found in the specified models for the kb subset
            symbol_set = set()
            for model in models:
                unique_symbols_in_model = set()
                reconstructed_kb[node]["models_kb"][model] = node_kb["models_kb"][model]
                for event in reconstructed_kb[node]["models_kb"][model]['sequence']:
                    for sym in event:
                        symbol_set.add(sym)
                        unique_symbols_in_model.add(sym)
                    symbol_freq_counter.update(event)

                symbol_model_member_freq_counter.update(unique_symbols_in_model)

            for sym in symbol_set:
                reconstructed_kb[node]['symbols_kb'][sym] = node_kb['symbols_kb'][sym]

                # update frequency and model member frequency to new subset values
                reconstructed_kb[node]['symbols_kb'][sym]['features']['frequency'] = symbol_freq_counter[sym]
                reconstructed_kb[node]['symbols_kb'][sym]['features']['model_member_frequency'] = symbol_model_member_freq_counter[sym]

                # if the symbol is a vector, add it to the subset vectors_kb
                if 'VECTOR|' in sym:
                    vectHash = sym.split('|')[-1]
                    reconstructed_kb[node]['vectors_kb'][vectHash] = node_kb['vectors_kb'][vectHash]

        return reconstructed_kb
    except Exception as e:
        print(f'failed to retrieve kb subset: {e}')
        raise e


def recursive_delete_model(agent: AgentClient, model_name: str, nodes=None, log_to_file: bool = False):
    """Recursively remove a model from an Agent.

    Args:
        agent (AgentClient): GAIuS Agent
        model_name (str): Model to remove
        nodes (_type_, optional): Nodes to delete model from. Defaults to None.
        log_to_file (bool, optional): output log data to file. Defaults to False.
    """

    if not agent._connected:
        return "Agent not connected"

    if log_to_file:
        import time
        t = time.localtime()
        file = open(f"{time.strftime('recursive_model_delete_log_%Y_%m_%d_%H_%M_%S', t)}_log.txt", mode='w+')
    
    # variables
    changed_models = {}
    primitve_patterns_to_remove = []
    symbols_to_update_in_other_nodes = {}
    
    if nodes is None:
        nodes = agent.all_nodes
    else:
        nodes = [node for node in agent.all_nodes if (node['name'] in nodes)]

    overall_result = {}
    
    for node in nodes:
        changed_models[node["name"]] = dict()
    
    for node in nodes:
        try:
            # call delete model on the passed nodes
            response = agent.delete_model(model_name, [node["name"]])
            overall_result[node["name"]] = response[node["name"]]
            
            if(log_to_file):
                file.write(f"initial delete_model response = {response}")
            
            # if model not in node, need to look for patterns
            # else add the pattern with the passed model name to primitve_patterns_to_remove 
            if(response[node["name"]] != "deleted"):
                if(log_to_file):
                    file.write("Trying to recursive delete models")
                for pattern in primitve_patterns_to_remove:
                    if(log_to_file):
                        file.write(f"Looking for pattern {pattern} in {node['name']}")
                    result = agent.remove_patterns_from_system([pattern])

                    # go through nodes again to make sure that pattern gets removed
#                         print(f"result = {result}")
#                         print(f"result.keys() = {result.keys()}")
                    if result == {}:
                        continue

                    for node1 in nodes:
                        if(log_to_file):
                            file.write(f"node1 = {node1['name']}")
                        # if nothing was found, then go to next pattern
                        if(log_to_file):
                            file.write(f"result[{node1['name']}] = {result[node1['name']] if (result[node1['name']] == {}) else result[node1['name']]['models']}")
                        
                        if(result[node1["name"]] == {}):
                            continue

                        for old_model_hash, new_model_hash in result[node1["name"]]["models"].items():
                            # so if that caused a model to be deleted, add this model
                            # to list of patterns to remove

                            # else update models further 
                            if(new_model_hash == "deleted"):
                                if(log_to_file):
                                    file.write(f"Adding .+\|{node1['id']}\|{old_model_hash}.+ to primitve_patterns_to_remove")
                                
                                pattern = f".+\|{node1['id']}\|{old_model_hash}.+"

                                # don't add the pattern if its already there
                                if pattern in primitve_patterns_to_remove:
                                    continue

                                primitve_patterns_to_remove.append(f".+\|{node1['id']}\|{old_model_hash}.+")
                                changed_models[node1["name"]][old_model_hash] = "deleted"
                            else:
                                if(log_to_file):
                                    file.write(f"Adding .+\|{node1['id']}\|{old_model_hash}.+ to symbols_to_update_in_other_nodes")
                                pattern = f".+{old_model_hash}.+"
                                # don't add the pattern if its already there
                                if pattern in symbols_to_update_in_other_nodes:
                                    continue

                                symbols_to_update_in_other_nodes[f".+{old_model_hash}.+"] = new_model_hash
                                changed_models[node1["name"]][old_model_hash] = new_model_hash
                    
                if(log_to_file):
                    file.write(f"Updating symbols")
                
                for old_symbol, new_symbol in symbols_to_update_in_other_nodes.items():
                    '''
                    Steps:
                        1) Get models which contain an old symbol
                        2) Get their sequences
                        3) Replace the symbol in their sequences with the new one
                        4) Delete old model
                        5) Add new model
                        6) Add record to symbols_to_update_in_other_nodes with old_model
                    '''
                    # get all models with pattern that needs to be updated
                    models_with_pattern = agent.get_models_with_patterns([old_symbol])

                    for node1 in nodes:
                        if (log_to_file):
                            file.write(f"models_with_pattern[{node1['name']}] = {models_with_pattern[node1['name']]} with {old_symbol}")

                        if (models_with_pattern[node1["name"]]["model_list"] == []):
                            continue

                        if (log_to_file):
                            file.write("node1['name'] not an empty list")

                        # go through each of them and delete the old model
                        # and add a model with the new symbol
                        for model in models_with_pattern[node1["name"]]["model_list"]:
                            model_instance = agent.get_model(model, nodes=[node1['name']])
                            if model_instance[node1["name"]] == {}:
                                continue


                            model_sequence = model_instance[node1["name"]]["sequence"]
                            model_frequency = model_instance[node1["name"]]["frequency"]
                            model_emotives = model_instance[node1["name"]]["emotives"]

                            # delete the model
                            response = agent.delete_model(model)
                            if (log_to_file):
                                file.write(f"response to delete_model[{node1['name']}] = {response[node1['name']]} for model = {model}")

                            # don't continue unless it was actually deleted
                            if(response[node1["name"]] != "deleted"):
                                if (log_to_file):
                                    file.write(f"{response[node1['name']]}")
                                continue
                            else:
                                if (log_to_file):
                                    file.write(f"delete old model response = {response}")

                            # replace patterns
                            pure_hash = old_symbol
                            pure_hash = pure_hash.replace(".+", "")
                            for i in range(0, len(model_sequence)):
                                for j in range(0, len(model_sequence[i])):
                                    if (pure_hash in model_sequence[i][j]):
                                        if (log_to_file):
                                            file.write(f"replaced {model_sequence[i][j]} with {new_symbol}")
                                        model_sequence[i][j] = model_sequence[i][j].replace(pure_hash, new_symbol)

                            # add model with new symbols
                            response = agent.add_model(sequence = model_sequence,
                                            vector=[],
                                            frequency = model_frequency,
                                            emotives = model_emotives,
                                            nodes=[node1['name']])

                            if (log_to_file):
                                file.write(f"response to add_model = {response}")

                            changed_models[node1["name"]][model] = response[node1["name"]]["name"]

            else:
                if (log_to_file):
                    file.write(f"Added .+\|{node['id']}\|{model_name}.+ pattern to primitve_patterns_to_remove")
                primitve_patterns_to_remove.append(f".+\|{node['id']}\|{model_name}.+")
        
        except Exception as exception:
            raise str(exception)

    if (log_to_file):
        file.close()

    # TODO changed models can be returned, but becomes super messy if you update a model several times
    return overall_result, changed_models


def recursive_update_model(agent: AgentClient, model_name: str,
                           model: dict = {}, nodes = None, log_to_file: bool = False):
    """Recursively update model in a hierarchical agent

    Args:
        agent (AgentClient): GAIuS Agent
        model_name (str): Model name to update
        model (dict): updated model contents
        nodes (_type_, optional): _description_. Defaults to None.
        log_to_file (bool, optional): _description_. Defaults to False.
    """

    if (not agent._connected):
        return "Agent not connected"

    if (log_to_file):
        import time
        t = time.localtime()
        file = open(f"{time.strftime('recursive_model_update_log_%Y_%m_%d_%H_%M_%S', t)}_log.txt", mode='w+')

    # variables
    changed_models = {}
    primitve_patterns_update_in_other_nodes = {}

    if nodes is None:
        nodes = agent.all_nodes
    else:
        nodes = [node for node in agent.all_nodes if (node['name'] in nodes)]

    overall_result = {}

    for node in nodes:
        changed_models[node["name"]] = dict()

    for node in nodes:
        try:
            # call delete model on the passed nodes
            response = agent.update_model(model_name, model, [node["name"]])
            overall_result[node["name"]] = response[node["name"]]

            if (log_to_file):
                file.write(f"initial update_model response = {response[node['name']]}\n")

            # if model was not in node, just skip
            if (response[node['name']].find("failed") != -1):
                continue

            changed_models[node["name"]] = response[node["name"]]

            # if model_hash did not change, we can simple add it to the list of symbols to update
            file.write(f"{model=}")
            model["name"] = response[node['name']]
            primitve_patterns_update_in_other_nodes[f".+\|{node['id']}\|{model_name}.+"] = model

            if (log_to_file):
                file.write("Trying to recursive update models\n")

            for node1 in nodes:
                # current node won't contain a pattern its updated
                if (node1 == node):
                    continue

                patterns_to_add = {}
                for pattern, pattern_data in primitve_patterns_update_in_other_nodes.items():
                    if (log_to_file):
                        file.write(f"Looking for pattern {pattern} in {node1['name']}\n")
                    # need to get all models that contain this model in their sequences
                    result = agent.get_models_with_patterns([pattern], [node1["name"]])[node1["name"]]["model_list"]

                    if (log_to_file):
                        file.write(f"Result for {pattern} in {node1['name']} {result=}\n")

                    if (result == []):
                        continue

                    model_name_to_find = pattern.split("|")[-1][0:-2]
                    if (log_to_file):
                        file.write(f"{model_name_to_find=}")

                    # for each model that needs to be updated, we update
                    for model_name in result:
                        # get model data
                        model_data = agent.get_model(model_name)[node1["name"]]

                        # find where the model name appears and replace it
                        for i in range(len(model_data["sequence"])):
                            for j in range(len(model_data['sequence'][i])):
                                if (model_data["sequence"][i][j].find(model_name_to_find) != -1):
                                    model_data["sequence"][i][j] = \
                                        model_data["sequence"][i][j].replace(model_name_to_find, pattern_data["name"])

                        # convert emotives in model_data and average it
                        for key, item in model_data["emotives"].items():
                            emotive_data = model_data["emotives"][key]
                            if not (type(emotive_data) is list):
                                continue

                            model_data["emotives"][key] = sum(emotive_data) / len(emotive_data)

                        # replace emotive values (no way to really average it)
                        for key, item in pattern_data["emotives"].items():
                            model_data["emotives"][key] = item

                        # replace metadata
                        model_data["metadata"] = pattern_data["metadata"]

                        # replace pattern with new data
                        recur_result = agent.update_model(model, model_data, [node1["name"]])[node1["name"]]

                        # add updated pattern
                        changed_models[node1["name"]][model_data["name"]] = recur_result
                        model_data["name"] = recur_result
                        if (pattern in patterns_to_add):
                            continue
                        patterns_to_add[f".+\|{node1['id']}\|{model_name}.+"] = model_data

                for pattern, pattern_data in patterns_to_add.items():
                    # add updated pattern
                    if (pattern in primitve_patterns_update_in_other_nodes):
                        continue
                    primitve_patterns_update_in_other_nodes[pattern] = pattern_data

        except Exception as exception:
            raise str(exception)

    if (log_to_file):
        file.close()

    # TODO changed models can be returned, but becomes super messy if you update a model several times
    return overall_result, changed_models


def is_abstracted_symbol(symbol: str) -> bool:
    """Check if the symbol provided is an abstracted symbol

    Args:
        symbol (str): symbol

    Returns:
        bool: True or False
    """
    symbol_split = symbol.split('|', maxsplit=4)
    
    if len(symbol_split) < 4:
        return False
    
    if symbol_split[0] != 'PRIMITIVE':
        return False

    # all abstracted symbols reference primitive id
    if not symbol_split[1].startswith('p'):
        return False

    if not symbol_split[3] in ['matches', 'missing', 'extras', 'present', 'past', 'future', 'classification', 'name']:
        return False

    return True

def remove_abstracted_symbols(agent: AgentClient, symbols: list):
    """Take an abstracted symbol, split it into corresponding parts, and call remove symbol
    on the symbol that has been abstracted up.
    
    Symbol format: PRIMITIVE|<NODE_ID>|<MODEL HASH>|<ABSTRACTED FIELD>|<ABSTRACTED VALUE>

    Args:
        agent (AgentClient): Agent to perform removal on
        symbols (list): Abstracted symbols to be removed.

    """
    models_to_delete = defaultdict(list)
    symbols_to_delete = defaultdict(list)
    for symbol in symbols:
        if not is_abstracted_symbol(symbol=symbol):
            print(f"provided symbol {symbol} is not an abstracted symbol")
            continue

        symbol_split = symbol.split('|', maxsplit=4)

        if symbol_split[0] != 'PRIMITIVE':
            raise Exception(f'Symbol {symbol} does not begin with "PRIMITIVE|"')

        node_id = symbol_split[1]
        model_hash = symbol_split[2]
        abstracted_field = symbol_split[3]
        abstracted_value = symbol_split[4]

        # escaped_abstracted_value = re.escape(abstracted_value)
        # escaped_original_value = re.escape(symbol)

        node_name = agent.genome.primitives[node_id]["name"]

        symbols_to_delete['all'].append(symbol)

        if abstracted_field == 'name':
            models_to_delete[node_name].append(abstracted_value)
        else:
            symbols_to_delete[node_name].append(abstracted_value)
    
    for node_name, model_list in models_to_delete.items():
        print(f'deleting {len(model_list)} models from node {node_name}:')
        for model in model_list:
            agent.delete_model(model_name=model, nodes=node_name)
    
    for node_name, symbol_list in symbols_to_delete.items():
        if node_name == 'all':
            print(f'removing symbols {symbol_list} from all nodes')
            agent.remove_symbols_from_system(symbols_list=symbol_list)
            continue
        print(f'removing symbols {symbol_list} from node {node_name}')
        agent.remove_symbols_from_system(symbols_list=symbol_list, nodes=node_name)

    return

def merge_single_node_kb(kb1: Dict[str, dict], kb2: Dict[str, dict]) -> dict:
    """Merge two single node KBs. Used internally by merge_kbs

    Args:
        kb1 (Dict[str, dict])
        kb2 (Dict[str, dict])

    Raises:
        Exception: _description_

    Returns:
        dict: _description_
    """
    result_kb = {'models_kb': {},
                 'vectors_kb': {},
                 'symbols_kb': {},
                 'metadata': {}
                 }

    fields = ['symbols_kb', 'metadata', 'vectors_kb', 'models_kb']
    assert all(field in kb1 for field in fields)
    assert all(field in kb2 for field in fields)

    # ensure KB models are completely disjoint,
    # otherwise raise exception and kick back to user for manual resolution
    kb1_models = list(kb1['models_kb'].keys())
    if any(model in kb2['models_kb'] for model in kb1_models):
        overlapping_models = set(kb1_models)
        overlapping_models.intersection_update(kb2['models_kb'].keys())
        raise Exception(
            f'Overlapping models between kbs need manual resolution: {overlapping_models=}')

    # get symbol frequencies when iterating through models
    symbol_freq_counter = Counter()
    symbol_model_member_freq_counter = Counter()
    total_model_frequencies = 0
    symbol_set = set()

    for model_name, model in kb1['models_kb'].items():
        unique_symbols_in_model = set()
        result_kb["models_kb"][model_name] = model
        for event in model['sequence']:
            for sym in event:
                symbol_set.add(sym)
                unique_symbols_in_model.add(sym)
            symbol_freq_counter.update(event)
        symbol_model_member_freq_counter.update(unique_symbols_in_model)
        total_model_frequencies += model['frequency']

    for model_name, model in kb2['models_kb'].items():
        unique_symbols_in_model = set()
        result_kb["models_kb"][model_name] = model
        for event in model['sequence']:
            for sym in event:
                symbol_set.add(sym)
                unique_symbols_in_model.add(sym)
            symbol_freq_counter.update(event)

        symbol_model_member_freq_counter.update(unique_symbols_in_model)
        total_model_frequencies += model['frequency']

    sym: str
    for sym in symbol_set:
        result_kb['symbols_kb'][sym] = {'name': sym,
                                        'features': {}}

        # update frequency and model member frequency to merged values
        result_kb['symbols_kb'][sym]['features']['frequency'] = symbol_freq_counter[sym]
        result_kb['symbols_kb'][sym]['features']['model_member_frequency'] = symbol_model_member_freq_counter[sym]

        # if the symbol is a vector, add it to the merged vectors_kb
        if 'VECTOR|' in sym:
            vectHash = sym.split('|')[-1]
            result_kb['vectors_kb'][vectHash] = kb1['vectors_kb'].get(
                vectHash, kb2['vectors_kb'].get(vectHash))

    result_kb['metadata'] = {
        "total_model_frequencies": total_model_frequencies,
        "total_symbol_frequencies": sum(symbol_freq_counter.values()),
        "total_symbols_in_models_frequencies": sum(symbol_model_member_freq_counter.values())
    }
    return result_kb

def merge_kbs(kb1: dict, kb2: dict) -> Dict[str, dict]:
    """Merge two Knowledgebases

    can perform merging on two kbs with multiple nodes, or a single node kb

    Args:
        kb1 (dict)
        kb2 (dict)

    Raises:
        Exception: If there are multiple nodes in the kb, and they do not line up exactly, raise Exception

    Returns:
        Dict[str, dict]: merged KB
    """
    # print(f'KB1={kb1}')
    resultant_kb = {}
    if 'metadata' in kb1:
        resultant_kb = merge_single_node_kb(kb1=kb1, kb2=kb2)
        return resultant_kb

    # multi-node kb merging
    if sorted(kb1.keys()) != sorted(kb2.keys()):
        raise Exception(
            f"Nodes mismatch between KB1 {list(kb1.keys())} and KB2 {list(kb2.keys())}")
    for node in kb1:
        resultant_kb[node] = merge_single_node_kb(kb1=kb1[node], kb2=kb2[node])
    return resultant_kb



import functools
import json
import uuid
import traceback
from typing import Dict, List, Any, Tuple, Union
from os.path import join, splitext
from datetime import datetime
from copy import deepcopy
import requests
from zipfile import ZipFile, ZIP_DEFLATED, is_zipfile
import warnings
from ia.gaius.genome_info import Genome

class AgentQueryError(BaseException):
    """Raised if any query to any node returns an error."""
    pass


class AgentConnectionError(BaseException):
    """Raised if connecting to any node returns an error."""
    pass

class AgentQueryWarning(Warning):
    """Used to warn user that a query partially did not complete"""

def _ensure_connected(f):
    @functools.wraps(f)
    def inner(self, *args, **kwargs):
        if not self._connected:
            raise AgentConnectionError(
                'Not connected to a agent. You must call `connect()` \
                    on a AgentClient instance before making queries'
            )
        return f(self, *args, **kwargs)

    return inner


def _remove_unique_id(response: dict) -> dict:
    """Return *response* with the key 'unique_id'
        removed regardless of nesting."""
    if isinstance(response, dict):
        if 'unique_id' in response:
            del response['unique_id']
        for value in response.values():
            if isinstance(value, dict):
                _remove_unique_id(value)
    return response


class AgentClient:
    """Interface for interacting with agents."""

    def __init__(self,
                 bottle_info: dict,
                 verify: bool = True,
                 timeout: float = None):
        """
        Provide agent information in a dictionary.

        Args:
            bottle_info (dict, required): the info used to connect and
                authenticate with a GAIuS agent
            verify (bool, optional): whether to enable SSL Verification
                on api commands to an agent

        Example:
            .. code-block:: python

                from ia.gaius.agent_client import AgentClient

                agent_info = {'api_key': 'ABCD-1234',
                            'name': 'gaius-agent',
                            'domain': 'intelligent-artifacts.com',
                            'secure': False}

                agent = AgentClient(agent_info)
                agent.connect()

                agent.set_ingress_nodes(['P1'])
                agent.set_query_nodes(['P1'])

        """
        self.session = requests.Session()
        self.session.request = functools.partial(self.session.request, timeout=timeout)
        self._timeout = timeout
        self.genome = None
        self._bottle_info = bottle_info
        self.name = bottle_info['name']
        self._domain = bottle_info['domain']
        self._api_key = bottle_info['api_key']
        self.ingress_nodes = []
        self.query_nodes = []
        self._headers = {'X-API-KEY': self._api_key}
        self.all_nodes = []
        self._connected = False
        self.genome = None
        self.gaius_agent = None
        self.send_unique_ids = True
        self.summarize_for_single_node = True
        self._verify = verify
        if 'secure' not in self._bottle_info or self._bottle_info['secure']:
            self._secure = True
            if not self.name:
                self._url = 'https://{domain}/'.format(**self._bottle_info)
            else:
                self._url = 'https://{name}.{domain}/'.format(**self._bottle_info)
        else:
            self._secure = False
            if not self.name:
                self._url = 'http://{domain}/'.format(**self._bottle_info)
            else:
                self._url = 'http://{name}.{domain}/'.format(**self._bottle_info)

    def __repr__(self) -> str:
        return (
            '<{name}.{domain}| secure: %r, connected: %s, gaius_agent: %s, \
                  ingress_nodes: %i, query_nodes: %i>'.format(
                **self._bottle_info
            )
            % (
                self._secure,
                self._connected,
                self.gaius_agent,
                len(self.ingress_nodes),
                len(self.query_nodes),
            )
        )

    def get_interface_node_config(self, interface_nodes: list) -> dict:
        """Produce copy of agent config with interface_node field for use with Thinkflux

        Args:
            interface_nodes (list): nodes to use as interface nodes
        """
        interface_dict = deepcopy(self._bottle_info)
        interface_dict['interface_nodes'] = interface_nodes
        
        return interface_dict
    
    def set_timeout(self, timeout: float):
        """Redefine timeout for AgentClient API Calls. Requires re-instantiating requests.Session object

        Args:
            timeout (float): The requested timeout (in seconds)
        """
        self.session = requests.Session()
        self.session.request = functools.partial(self.session.request, timeout=timeout)
        return
    
    def receive_unique_ids(self, should_set: bool = True) -> bool:
        self.send_unique_ids = should_set
        return self.send_unique_ids

    def connect(self) -> Dict:
        """Establishes initial connection to GAIuS agent and grabs the
            agent's genome for node definitions.

            Example:
                    .. code-block:: python

                        agent_info = {...}
                        agent = AgentClient(agent_info)
                        agent.connect()
        """
        response_data = self.session.get(f'{self._url}connect',
                                         verify=self._verify,
                                         headers=self._headers).json()
        if 'status' not in response_data or response_data['status'] != 'okay':
            self._connected = False
            raise AgentConnectionError("Connection failed!", response_data)

        self.genome = Genome(response_data['genome'])
        self.gaius_agent = response_data['genome']['agent']
        self.all_nodes = [{"name": i['name'], "id": i['id']} for i in self.genome.primitives.values()]
        if response_data['connection'] == 'okay':
            self._connected = True
        else:
            self._connected = False

        return {'connection': response_data['connection'],
                'agent': response_data['genie']}

    def set_ingress_nodes(self, nodes: List = None) -> List:
        """List of primitive names to define where data will be sent.

            Example:
                .. code-block:: python

                    # when observing data, it will be sent to only P1 by default
                    agent.set_ingress_nodes(nodes=["P1"])
        """
        if nodes is None:
            nodes = []
        self.ingress_nodes = [{'id': self.genome.primitive_map[node],
                               'name': node}
                              for node in nodes]
        return self.ingress_nodes

    def set_query_nodes(self, nodes: List = None) -> List:
        """List of primitive names to define which nodes should return answers.

            Example:
                .. code-block:: python

                    # when getting predictions, it will be received only from P1 by default
                    agent.set_query_nodes(nodes=["P1"])
        """
        if nodes is None:
            nodes = []
        self.query_nodes = [{'id': self.genome.primitive_map[node], 'name': node} for node in nodes]
        return self.query_nodes

    def _query(self,
               query_method: Any,
               path: str,
               data: Union[dict, str] = None,
               nodes: List = None,
               unique_id: str = None) -> Union[dict, Tuple[dict, str]]:
        """Internal helper function to make a REST API call with the given *query* and *data*."""
        if not self._connected:
            raise AgentConnectionError(
                'Not connected to a agent. You must call `connect()` on a AgentClient instance before making queries'
            )
        result = {}
        if unique_id is not None:
            if data:
                data['unique_id'] = unique_id
            else:
                data = {'unique_id': unique_id}

        data = json.loads(json.dumps(data))

        if isinstance(nodes[0], str):
            nodes = [{'name': name, 'id': self.genome.primitive_map[name]} for name in nodes]
        for node in nodes:
            full_path = f'{self._url}{node["id"]}/{path}'
            try:
                if data is not None:
                    response = query_method(full_path,
                                            verify=self._verify,
                                            headers=self._headers,
                                            json={'data': data})
                else:
                    response = query_method(full_path,
                                            verify=self._verify,
                                            headers=self._headers)
                response.raise_for_status()
                response = response.json()
                if response['status'] != 'okay':
                    raise AgentQueryError(response['message'])
                if not self.send_unique_ids:
                    response = _remove_unique_id(response['message'])
                else:
                    response = response['message']
                if len(nodes) == 1 and self.summarize_for_single_node:
                    result = response
                else:
                    result[node['name']] = response
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None
        if unique_id is not None:
            return result, unique_id
        return result

    def set_summarize_for_single_node(self, value: bool):
        """When True, queries against a single node return responses directly
        instead of in a dict key.

        Example:
            .. code-block:: python

                # set summarize_for_single_node to True
                >>> agent.set_summarize_for_single_node(value=True)
                >>> gdf={'strings':['hello'], 'vectors': [], 'emotives':{}}
                >>> agent.observe(gdf)
                'observed'

                # set summarize_for_single_node to False
                >>> agent.set_summarize_for_single_node(value=False)
                >>> agent.observe(gdf)
                {'P1': 'observed'}

        """
        self.summarize_for_single_node = value

    def observe(self, data: Dict, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Exclusively uses the 'observe' call.
            Will observe on nodes specified by set_ingress_nodes()
            unless nodes explicitly provided

            Args:
                data (dict): the GDF data to be observed
                nodes (list, optional): Nodes to observe data on,
                    defaults to ingress nodes

            Example:
                .. code-block:: python

                    from ia.gaius.utils import create_gdf
                    >>> gdf = create_gdf(strings=['hello'])
                    >>> agent.observe(data=gdf, nodes=["P1"])
                    'observed'

        """
        if nodes is None:
            nodes = self.ingress_nodes
        return self._query(self.session.post, 'observe', data=data, nodes=nodes)

    def _observe_event(self, data: Dict, unique_id: str = None) -> Tuple[dict, str]:
        """Exclusively uses the 'observe' call."""
        results = {}
        uid = None
        if unique_id is None:
            unique_id = str(uuid.uuid4())
        for node, node_data in data.items():
            response, uid = self._query(self.session.post,
                                        'observe',
                                        data=node_data,
                                        nodes=[node],
                                        unique_id=unique_id)
            results[node] = response
        return results, uid

    @_ensure_connected
    def observe_classification(self, data=None, nodes: List = None):
        """Send a classification to all nodes as a singular symbol in the
        last event.

        Sending the classification as a single symbol in the last event is
        the canonical way to classify a sequence.
        """
        if nodes is None:
            nodes = self.query_nodes
        return self._query(self.session.post, 'observe', data=data, nodes=nodes)

    def show_status(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Return the current status of the agent.

        Example:
            .. code-block:: python

                >>> agent.show_status(nodes=['P1'])
                {'PREDICT': True,
                'SLEEPING': False,
                'emotives': {},
                'last_learned_model_name': '',
                'models_kb': '{KB| objects: 0}',
                'name': 'P1',
                'size_WM': 0,
                'target': '',
                'time': 0,
                'vectors_kb': '{KB| objects: 0}'}
        """
        if nodes is None:
            nodes = self.all_nodes
        return self._query(self.session.get, 'status', nodes=nodes)

    def learn(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Return the learn results.

        Example:
            .. code-block:: python

                >>> agent.learn(nodes=["P1"])
                'MODEL|004da01b0ef40d7c3e0965a6b4fd0f413672fbff'
        """
        if nodes is None:
            nodes = self.all_nodes
        return self._query(self.session.post, 'learn', nodes=nodes)

    def get_wm(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Return information about Working Memory.

        Example:
            .. code-block:: python

                >>> gdf = {'strings':['hello'], 'vectors': [], 'emotives':{}}
                >>> agent.observe(gdf, nodes=['P1'])
                >>> agent.get_wm(nodes=['P1'])
                [['hello']]

        """
        if nodes is None:
            nodes = self.all_nodes
        return self._query(self.session.get, 'working-memory', nodes=nodes)

    def get_predictions(self, unique_id: str = None, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Return prediction result data.

        Example:
            .. code-block:: python

                predictions = agent.get_predictions(nodes=["P1"])
        """
        if nodes is None:
            nodes = self.query_nodes
        return self._query(self.session.post, 'predictions', nodes=nodes, unique_id=unique_id)

    def clear_wm(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Clear the Working Memory of the Agent.
        Will not delete any persisted memory from the KnowledgeBase

        Example:
            .. code-block:: python

                >>> agent.clear_wm(nodes=['P1'])
                'wm-cleared'
        """
        if nodes is None:
            nodes = self.all_nodes
        return self._query(self.session.post, 'working-memory/clear', nodes=nodes)

    def clear_all_memory(self, clear_blacklist=True, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Clear both the Working Memory and persisted memory.
        Equivalent to re-initializing the nodes specified

        Args:
            clear_blacklist (bool, optional): Controls if symbol blacklist is
                also cleared or not. Defaults to False.
            nodes (List, optional): List of nodes to call command on.
                Defaults to None.

        .. code-block:: python

                >>> agent.clear_all_memory(nodes=["P1"])
                'all-cleared'

        Returns:
            Union[dict, Tuple[dict, str]]: _description_
        """
        if nodes is None:
            nodes = self.all_nodes

        data = {"clear_blacklist": clear_blacklist}

        return self._query(self.session.post, 'clear-all-memory', nodes=nodes, data=data)

    def get_percept_data(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Return percept data.

        Example:
            .. code-block:: python

                percept_data = agent.get_percept_data(nodes=["P1"])
        """
        if nodes is None:
            nodes = self.all_nodes
        return self._query(self.session.get, 'percept-data', nodes=nodes)

    def get_cognition_data(self, unique_id: str = None, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Return cognition data.

        Example:
            .. code-block:: python

                cog_data = agent.get_cognition_data(nodes=["P1"])
        """
        if nodes is None:
            nodes = self.query_nodes
        return self._query(self.session.get, 'cognition-data', nodes=nodes, unique_id=unique_id)

    def get_all_genes(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Return all genes for nodes.

        Example:
            .. code-block:: python

                genes = agent.get_all_genes(nodes=["P1"])
        """
        if nodes is None:
            nodes = self.all_nodes
        return self._query(self.session.get, 'all-genes', nodes=nodes)

    @_ensure_connected
    def change_genes(self, gene_data: Dict, nodes: List = None) -> Union[dict, Any]:
        """Change the genes in *gene_data* to their associated values.

        This will do live updates to an existing agent, rather than stopping
        an agent and starting a new one, as per 'injectGenome'.
        gene_data of form:

            {gene: value}

        Example:
            .. code-block:: python

                agent.change_genes(gene_data={"recall_threshold" : 0.15})

        """
        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]

        result = {}
        for node in nodes:
            response = self.session.post(
                f"{self._url}{node['id']}/genes/change",
                verify=self._verify,
                headers=self._headers,
                json={'data': gene_data},
            ).json()
            if 'error' in response or response['status'] == 'failed':
                if len(nodes) == 1 and self.summarize_for_single_node:
                    raise AgentQueryError(response)
            self.genome.change_genes(node['id'], gene_data)
            if len(nodes) == 1 and self.summarize_for_single_node:
                return response['message']
            result[node['name']] = response['message']
        return result

    @_ensure_connected
    def get_gene(self, gene: str, nodes: List = None) -> Union[dict, Dict[Any, Dict[str, Any]]]:
        """Return the value for the gene *gene* on *nodes*.

        Example:
                .. code-block:: python

                    agent.get_gene(gene="recall_threshold")

        """
        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.get(f"{self._url}{node['id']}/gene/{gene}",
                                            verify=self._verify,
                                            headers=self._headers
                                            ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return {gene: response['message']}
                result[node['name']] = {gene: response['message']}
            except BaseException as exception:
                raise AgentQueryError(exception) from None

        return result

    @_ensure_connected
    def get_model(self, model_name: str, nodes: List = None) -> Union[dict, Any]:
        """Returns model with name *model_name*.

        Model name is unique, so it should not matter that we query all nodes, only
        one model will be found.

        Args:
            model_name (str): Name of model to retrieve
            nodes (List, optional): list of nodes to retrieve model from.
                Defaults to None (all nodes).

        Returns:
            Union[dict, Any]: dictionary of {node_name : model} pairs
        Example:
                .. code-block:: python

                    agent.get_model("MODEL|d85f28ebda228064299035bf2dbf58f5c6484108")

        """
        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.get(f"{self._url}{node['id']}/model/{model_name}",
                                            headers=self._headers,
                                            verify=self._verify
                                            ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return response['message']
                else:
                    result[node['name']] = response['message']
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None

        return result

    @_ensure_connected
    def delete_model(self, model_name: str, nodes: List = None) -> Union[dict, Any]:
        """Deletes model with name *model_name*.

        Model name is unique, so it should not matter that we query all nodes,
        all its duplicates everywhere will be deleted.
        """
        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.delete(f"{self._url}{node['id']}/model/{model_name}",
                                               headers=self._headers,
                                               verify=self._verify
                                               ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return response['message']
                else:
                    result[node['name']] = response['message']
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None

        return result

    @_ensure_connected
    def update_model(self, model_name: str, model: Dict, nodes: List = None) -> Union[dict, Any]:
        """Update the model  **model_name** with the data in **model**
        on update_model **nodes**

        .. note::

            This function is only capable of updating the emotives
            corresponding to a model. In order to update the sequence
            corresponding to a model, delete this model and reobserve
            the updated sequence

        Example:
            .. code-block:: python

                # retrieve a model already learned on an agent
                model_name = 'MODEL|d85f28ebda228064299035bf2dbf58f5c6484108'
                model = agent.get_model(model_name=model_name, nodes=['P1'])

                # update the emotives in the model
                model["emotives"] = {'utility': [150, 200, 500]}
                agent.update_model(model_name=model_name, model=model, nodes=['P1'])

        """
        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.put(f"{self._url}{node['id']}/model/{model_name}",
                                            headers=self._headers,
                                            verify=self._verify,
                                            json={'model': model}
                                            ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return response['message']
                else:
                    result[node['name']] = response['message']
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None

        return result

    @_ensure_connected
    def resolve_model(self, model_name: str, nodes: List = None) -> Union[dict, Any]:
        """Returns model with name *model_name*.

        Model name is unique, so it should not matter that we query all nodes,
        only one model will be found.
        """
        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.get(f"{self._url}{node['id']}/model/{model_name}",
                                            headers=self._headers,
                                            verify=self._verify
                                            ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return response['message']
                else:
                    result[node['name']] = response['message']
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None

        return result

    def get_name(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Return name of *nodes*.

            Example:
                .. code-block:: python

                    agent.get_name(nodes=["P1", "P2"])

        """
        if nodes is None:
            nodes = self.all_nodes
        return self._query(self.session.get, 'name', nodes=nodes)

    def get_time(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Return time on *nodes*.

            Example:
                .. code-block:: python

                    >>> agent.get_time(nodes=['P1'])
                    '1'
        """
        if nodes is None:
            nodes = self.all_nodes
        return self._query(self.session.get, 'time', nodes=nodes)

    @_ensure_connected
    def get_models_with_symbols(self, symbols_list: List = None, nodes: List = None) -> Union[dict, Any]:
        '''
            Function which search the kbs on the nodes for models with certain
            symbols. All symbols must be present in model.sequence, or the
            model will not be returned
        '''
        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.post(f"{self._url}{node['id']}/get-models-with-symbols",
                                             verify=self._verify,
                                             headers=self._headers,
                                             json={'data': symbols_list}
                                             ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return response['message']
                else:
                    result[node['name']] = response['message']
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None
        return result

    @_ensure_connected
    def get_models_with_patterns(self,
                                 symbols_list: List = None,
                                 nodes: List = None) -> Union[dict, Any]:
        '''
            Function which search the kbs on the nodes for models with
            certain regex patterns.

            All regex patterns must be present in model.sequence, or
            the model will not be returned
        '''
        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.post(f"{self._url}{node['id']}/get-models-with-patterns",
                                             verify=self._verify,
                                             headers=self._headers,
                                             json={'data': symbols_list},
                                             ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return response['message']
                else:
                    result[node['name']] = response['message']
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None
        return result

    @_ensure_connected
    def add_model(self, sequence: List,
                  vector: List = None,
                  frequency: int = 1,
                  emotives: Dict = {},
                  metadata: List = [],
                  nodes: List = None) -> Union[dict, Any]:
        '''
            Function to explicitely add a model with the given sequence
            of symbols, frequency, and emotives. Returns the hash of
            the generated function
        '''
        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.put(f"{self._url}{node['id']}/add-model",
                                            verify=self._verify,
                                            headers=self._headers,
                                            json={'sequence': sequence,
                                                  'vector': vector,
                                                  'frequency': frequency,
                                                  'emotives': emotives,
                                                  "metadata": metadata},
                                            ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return response['message']
                else:
                    result[node['name']] = response['message']
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None
        return result

    @_ensure_connected
    def remove_symbols_from_system(self,
                                   symbols_list: List,
                                   nodes: List = None) -> Union[dict, Any]:
        """Function which search the kbs on the nodes for models with
        certain symbols.

        Symbols are removed, and the symbols are blacklisted, thus making
        it impossible fo them to be in the working memory in the future.

        Args:
            symbols_list (List): list of symbols to remove. Defaults to None.
            nodes (List, optional): list of nodes to remove symbols from.
            Defaults to None (all nodes).

        Raises:
            AgentQueryError: Error in Cognitive Processor handling

        Returns:
            Union[dict, Any]: dict showing old model hash -> new model hash
            (or 'deleted') pairs
        """
        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.delete(f"{self._url}{node['id']}/symbols/remove",
                                               verify=self._verify,
                                               headers=self._headers,
                                               json={'data': symbols_list},
                                               ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return response['message']
                else:
                    result[node['name']] = response['message']
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None
        return result

    @_ensure_connected
    def remove_patterns_from_system(self,
                                    symbols_list: List,
                                    nodes: List = None) -> Union[dict, Any]:
        """
        Function which search the kbs on the nodes for models with
        certain regex patterns. Removed symbols that matches patterns,
        but does not blacklist them. So patterns may show up in the future.

        Args:
            symbols_list (List): list of regex symbol patterns to remove.
            nodes (List, optional): list of nodes to remove symbols on.
                Defaults to None (all nodes).

        Raises:
            AgentQueryError: Error in Cognitive Processor handling

        Returns:
            Union[dict, Any]: dict showing old model hash -> new model hash
                (or 'deleted') pairs

        Example:
            .. code-block:: python

                from ia.gaius.kb_ops import list_symbols

                symbols_before = list_symbols(agent)
                agent.remove_patterns_from_system(['VECTOR|.+'])
                symbols_after = list_symbols(agent)

        """
        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.delete(f"{self._url}{node['id']}/patterns/remove",
                                               verify=self._verify,
                                               headers=self._headers,
                                               json={'data': symbols_list}
                                               ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return response['message']
                else:
                    result[node['name']] = response['message']
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None
        return result

    def add_blacklisted_symbols(self,
                                symbols_list: List,
                                nodes: List = None):
        """Add a symbol to the blacklist manually.
        List all symbols currently in the blacklist using :func:`list_blacklisted_symbols`.
        Symbols may be removed from this list by using :func:`remove_blacklisted_symbols`.

        Args:
            symbols_list (List): list of symbols to add to the blacklist
            nodes (List, optional): list of nodes to add blacklisted symbols on.
                Defaults to None (all nodes).

        Raises:
            AgentQueryError: Error in Cognitive Processor handling

        Returns:
            dict or str: Depicting result of adding blacklisted symbols to each CP in nodes
        """

        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.post(f"{self._url}{node['id']}/blacklisted-symbols/add",
                                             verify=self._verify,
                                             headers=self._headers,
                                             json={'data': symbols_list}
                                             ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return response['message']
                else:
                    result[node['name']] = response['message']
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None
        return result

    def list_blacklisted_symbols(self, nodes: List = None):
        """Display a list of all symbols currently 'blacklisted' on each
        node in *nodes*.

        Symbols enter this list by being removed from a Cognitive Processor
        via :func:`remove_symbols_from_system` or
        :func:`remove_patterns_from_system`.

        Symbols may be removed from this list by using
        :func:`remove_blacklisted_symbols`.

        Args:
            nodes (List, optional): list of nodes to show blacklisted symbols
                on. Defaults to None (all nodes).

        Example:
            .. code-block:: python

                from ia.gaius.kb_ops import list_symbols,
                    get_models_containing_symbol

                symbols_before = list_symbols(agent)
                blacklisted_symbols = agent.list_blacklisted_symbols()
                # ensure all blacklisted symbols do not appear in the system
                assert all([sym not in symbols_before for sym in blacklisted_symbols])

                # ensure no models contain the blacklisted symbols
                models_containing_symbol = get_models_containing_symbol(agent, symbol_set=set(blacklisted_symbols))
                assert all([len(val) == 0 for val in models_containing_symbol.values()])
        """

        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.get(f"{self._url}{node['id']}/list-blacklisted-symbols",
                                            verify=self._verify,
                                            headers=self._headers
                                            ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return response['message']
                else:
                    result[node['name']] = response['message']
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None
        return result

    def remove_blacklisted_symbols(self,
                                   symbols_to_remove: List,
                                   nodes: List = None):
        """Remove symbols from blacklist on each Cognitive
        Processor in *nodes*

        Args:
            symbols_to_remove (List): list of symbols to remove from
                blacklisted symbols
            nodes (List, optional): list of nodes to remove blacklisted
                symbols on. Defaults to None (all nodes).

        Raises:
            AgentQueryError: Error in Cognitive Processor handling

        Returns:
            str: 'removed-blacklisted-symbols' or
            'failed-to-remove-some-symbols-from-blacklist'
        """

        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.delete(f"{self._url}{node['id']}/blacklisted-symbols/remove",
                                               verify=self._verify,
                                               headers=self._headers,
                                               json={'data': symbols_to_remove}
                                               ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return response['message']
                else:
                    result[node['name']] = response['message']
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None
        return result

    @_ensure_connected
    def get_vector(self, vector_name: str, nodes: List = None) -> Union[dict, Dict[Any, Dict[str, Any]]]:
        """Return the vector with *vector_name* on *nodes* (it will be present on at most one).

            Example:
                .. code-block:: python

                    agent.get_vector(vector_name="VECTOR|0b7f2aac43ae8a0d94c58be08a5b8b0ab28079de")

        """
        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.get(
                    f"{self._url}{node['id']}/vector", headers=self._headers, verify=self._verify, json={'data': vector_name}
                ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return response['message']
                else:
                    result[node['name']] = response['message']
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None

        return result

    @_ensure_connected
    def increment_recall_threshold(self, increment: float, nodes: List = None) -> Dict[Any, Dict[str, Any]]:
        """Increment recall threshold by *increment* on *nodes*.

            Args:
                increment (float): value by which to increment the recall threshold by

            Example:
                .. code-block:: python

                    agent.increment_recall_threshold(increment=0.05, nodes=["P1"])

            .. note::

                This command with only update the recall threshold by a specified value.
                In order to set the recall threshold to a specific value,
                use :func:`change_genes` instead.
        """
        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.post(
                    f"{self._url}{node['id']}/gene/increment-recall-threshold",
                    verify=self._verify,
                    headers=self._headers,
                    json={'increment': increment},
                ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                else:
                    self.genome.primitives[node['id']]['recall_threshold'] += increment
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return {"recall_threshold": response['message']}
                result[node['name']] = {"recall_threshold": response['message']}
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None

        return result

    def start_sleeping(self,
                       nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Tells *nodes* to start sleeping.

            Example:
                    .. code-block:: python

                        # will disable observing new data on node P1
                        agent.start_sleeping(nodes=["P1"])
        """
        if nodes is None:
            nodes = self.all_nodes
        return self._query(self.session.post, 'sleeping/start', nodes=nodes)

    def stop_sleeping(self,
                      nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Wakes up sleeping *nodes*.

            Example:
                .. code-block:: python

                    # will enable observing new data on node P1
                    agent.stop_sleeping(nodes=["P1"])

        """
        if nodes is None:
            nodes = self.all_nodes
        return self._query(self.session.post, 'sleeping/stop', nodes=nodes)

    def start_predicting(self,
                         nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Tells *nodes* to start predicting.

            Example:
                .. code-block:: python

                    # will enable prediction generation on node P1
                    agent.start_predicting(nodes=["P1"])
        """
        if nodes is None:
            nodes = self.all_nodes
        return self._query(self.session.post, 'predicting/start', nodes=nodes)

    def stop_predicting(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Tells *nodes* to stop predicting.

        Useful for faster training, but abstracted nodes will not learn.

        Example:
            .. code-block:: python

                # will disable prediction generation on node P1
                agent.stop_predicting(nodes=["P1"])
        """
        if nodes is None:
            nodes = self.all_nodes
        return self._query(self.session.post, 'predicting/stop', nodes=nodes)

    def start_autolearning(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Tells *nodes* to start predicting.

            Example:
                .. code-block:: python

                    # will enable prediction generation on node P1
                    agent.start_predicting(nodes=["P1"])
        """
        if nodes is None:
            nodes = self.all_nodes
        return self._query(self.session.post, 'autolearning/start', nodes=nodes)

    def stop_autolearning(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Tells *nodes* to stop predicting.

        Useful for faster training, but abstracted nodes will not learn.

        Example:
            .. code-block:: python

                # will disable prediction generation on node P1
                agent.stop_predicting(nodes=["P1"])
        """
        if nodes is None:
            nodes = self.all_nodes
        return self._query(self.session.post, 'autolearning/stop', nodes=nodes)

    @_ensure_connected
    def ping(self, nodes: List = None) -> Union[dict, Any]:
        """Ping a node to ensure it's up.

            Example:

                .. code-block:: python

                    agent.ping(nodes=["P1"])
        """
        if nodes is None:
            return self.session.get(f'{self._url}gaius-api/ping',
                                    headers=self._headers).json()

        nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        results = {}
        for node in nodes:
            response = self.session.get(f"{self._url}{node['id']}/ping",
                                        verify=self._verify,
                                        headers=self._headers).json()
            if 'error' in response or response["status"] == 'failed':
                if len(nodes) == 1 and self.summarize_for_single_node:
                    raise AgentQueryError(response)
                warnings.warn(response["message"], AgentQueryWarning)
            if len(nodes) == 1 and self.summarize_for_single_node:
                return response['message']
            results[node['name']] = response["message"]
        return results

    @_ensure_connected
    def get_kbs(self, directory='./'): # pragma: no cover
        """Returns the KBs for the agent.
        This is a mongo-db and can be used to store or analyze locally.
        Choose the directory to save in with the directory keyword.
        Default is in './'.

        .. warning::

            .. deprecated:: 11.0 (Kehl)

            Use :func:`get_kbs_as_json` instead.

        """
        _headers = self._headers
        _headers['Content-Encoding'] = 'gzip'
        archive = self.session.get(f"{self._url}database",
                                   headers=_headers,
                                   json={},
                                   verify=self._verify
                                   )
        archive_file = join(directory,
                            f'{self.gaius_agent}-{self.name}-{datetime.now()}-kb.archive.gz')
        with open(archive_file, 'wb') as f:
            f.write(archive.content)
        return f"Saved as {archive_file}"


    @_ensure_connected
    def _get_kbs_as_json_compressed(self,
                                   nodes: List = None,
                                   directory='./',
                                   filename=None,
                                   ids=True) -> str:
        """Returns KBs of specified nodes in a zip file (separated by node)

        """
        todays_date_to_seconds = str(datetime.now().isoformat(timespec='seconds'))
        field = 'name'
        if ids:
            field = 'id'
        if filename is None:
            filename = join(directory, f'{self.gaius_agent}-{self.name}-{todays_date_to_seconds}-kb.zip')
        with ZipFile(filename, 'w', compression=ZIP_DEFLATED) as kb_zip:
            if nodes is None:
                nodes = self.all_nodes
            prev_summarize_state = self.summarize_for_single_node

            try:
                self.set_summarize_for_single_node(False)
                for node in nodes:
                    output = self.session.get(f'{self._url}{node["id"]}/get_kb/raw', headers=self._headers)
                    if output.status_code != 200:
                        raise Exception(f'failed to get kb from node {node["id"]}')
                    print(f'writing kb for node {node["name"]}')
                    with kb_zip.open(f'{node[field]}.json', 'w', force_zip64=True) as f:
                        f.write(output.content)
            except Exception as e:
                print(f'failed to get kb: {e}')
                raise e
            finally:
                self.set_summarize_for_single_node(prev_summarize_state)

        return filename

    @_ensure_connected
    def get_kbs_as_json(self,
                        nodes: List = None,
                        directory='./',
                        filename=None,
                        separated=False,
                        ids=True,
                        obj=False,
                        compressed=False) -> Union[List[str], dict]:
        """Returns KBs of specified nodes as JSON Objects

        Args:
            nodes (list, optional): Defaults to None.
            directory (str, optional): Defaults to './'.
            separated (bool, optional): store each KB separately.
                Defaults to False
            ids (bool, optional): use primitive_ids as keys for knowledgebase
                instead of node numbers. Defaults to True
            obj (bool, optional): whether to save the kbs to a filepath
                or a python object returned from the function
            compressed (bool, optional): whether to save the kbs to a zip file. Not compatible with
                "separated" or "obj" arguments
        Returns:
            Union[List(str), Dict]: List of files where KB is stored
            if obj = False, or dictionary where KB is to be
            stored if obj = True


        Example:
            .. code-block:: python

                # save the KBs in a specified directory, without using node IDs
                # will use node names instead (e.g. P1, P2, P3)
                kb_path = "../path/to/save/dir/"
                agent.get_kbs_from_json(directory=kb_path, separated=False, ids=False)

                #save the KBs to the active directory, using node IDs (e.g. p46514fa2, etc.)
                agent.get_kbs_from_json(ids=True)

                #save the KBs to a dictionary object, using node names
                kb = agent.get_kbs_from_json(ids=False, obj=True)

        """
        todays_date_to_seconds = str(datetime.now().isoformat(timespec='seconds'))
        if obj is True and separated is True:
            print("obj == True and separated == True is not yet supported")
            return None

        if compressed == True:
            return self._get_kbs_as_json_compressed(nodes=nodes, directory=directory, filename=filename, ids=ids)

        if nodes is None:
            nodes = self.all_nodes
        prev_summarize_state = self.summarize_for_single_node
        try:
            self.set_summarize_for_single_node(False)
            output = self._query(requests.get, 'get_kb', nodes=nodes)
        except Exception as e:
            print(f'failed to get kb: {e}')
            raise e
        finally:
            self.set_summarize_for_single_node(prev_summarize_state)

        filenames = []
        if isinstance(output, dict):
            altered_output = {}
            if ids is True and "metadata" not in output:
                for key in output.keys():
                    node_id = [node["id"] for node in self.all_nodes if node['name'] == key][0]
                    altered_output[node_id] = output[key]
            else:
                altered_output = output
            if separated:
                for key in altered_output.keys():
                    archive_file = join(directory, f'{self.gaius_agent}-{self.name}-{key}-{todays_date_to_seconds}-kb.json')
                    filenames.append(archive_file)
                    with open(archive_file, 'w') as f:
                        f.write(json.dumps({key: altered_output[key]}))
            else:
                archive_file = join(directory, f'{self.gaius_agent}-{self.name}-{todays_date_to_seconds}-kb.json')
                if filename:
                    archive_file = filename

                filenames.append(archive_file)

                # return kb as json object
                if obj is True:
                    return altered_output
                # otherwise write json file
                with open(archive_file, 'w') as f:
                    f.write(json.dumps(altered_output))

        return filenames

    def _load_compressed_kb(self, path):
        node_dict = {}
        for item in self.all_nodes:
            node_dict[item['name']] = item['id']
            node_dict[item['id']] = item['id']

        with ZipFile(path, 'r') as kb_zip:
            kb_names = kb_zip.namelist()
            for filename in kb_names:
                node_name = splitext(filename)[0]
                if node_name not in node_dict:
                    print(f'node {node_name} not found in topology, skipping')
                    continue
                with kb_zip.open(filename, 'r') as f:
                    prim_id = node_dict[node_name]
                    response = self.session.post(f'{self._url}{prim_id}/load_kb',
                                            headers={'X-API-KEY': self._api_key,
                                                    'content-type':'application/json'},
                                            data=f)
                    if response.status_code != 200:
                        print(f'error loading kb for node {prim_id}: {response.text}')
                        return 'failed', response.status_code
                    response = self.session.post(f'{self._url}{prim_id}/working-memory/clear',
                                                headers={'X-API-KEY': self._api_key})
                    if response.status_code != 200:
                        print(f'error loading kb for node {prim_id}: {response.text}')
                        return 'failed', response.status_code

                    print(f'loading node {prim_id} succeeded')
        return 'success'

    @_ensure_connected
    def load_kbs_from_json(self, path=None, obj=None):
        """Load KBs from a JSON file

        Args:
            path (str, optional): path to JSON file where KBs are stored.
            obj  (dict, optional): object to load KnowledgeBase from

        Returns:
            str: 'success' or 'failed'

        Example:
            .. code-block:: python

                kb_path = "../path/to/kb.json"
                agent.load_kbs_from_json(path=kb_path)

        """

        node_dict = {}
        for item in self.all_nodes:
            node_dict[item['name']] = item['id']
            node_dict[item['id']] = item['id']

        kb = None
        if path is not None:
            if is_zipfile(path):
                # load_compressed_kb here
                return self._load_compressed_kb(path)

            with open(path, 'r') as f:
                kb = json.load(f)
        elif obj is not None:
            kb = obj
        else:
            raise Exception("Must specify path or obj argument")

        # check if the kb is only for a single node (simple-topology edge case)
        if 'metadata' in kb.keys():
            kb = {'P1': kb}

        for key, value in kb.items():
            prim_id = key
            if prim_id in node_dict:
                prim_id = node_dict[prim_id]
            else:
                print(f'node {prim_id} not found in topology, skipping')
                continue
            print(f'loading node {prim_id} from kb for {key}')
            response = self.session.post(f'{self._url}{prim_id}/load_kb',
                                         headers={'X-API-KEY': self._api_key},
                                         json={'data': value})
            if response.status_code != 200:
                print(f'error loading kb for node {prim_id}: {response.text}')
                return 'failed', response.status_code
            print(f'loading node {prim_id} succeeded')

        return 'success'

    @_ensure_connected
    def load_kbs_from_json_iter(self, path):
        """Uses ijson library to load KBs more efficiently

        Args:
            path (str): path to JSON file where KBs are stored.


        Returns:
            str: 'success' or 'failed'

        Example:
            .. code-block:: python

                kb_path = "../path/to/kb.json"
                agent.load_kbs_from_json(path=kb_path)

        """
        import ijson

        node_dict = {}
        for item in self.all_nodes:
            node_dict[item['name']] = item['id']
            node_dict[item['id']] = item['id']

        with open(path, 'rb') as kb_file:
            for key, value in ijson.kvitems(kb_file, '', use_float=True):
                prim_id = key
                if key in ['symbols_kb', 'vectors_kb', 'models_kb', 'metadata']:
                    raise Exception(f'use load_kbs_from_json() for a simple node')
                if prim_id in node_dict:
                    prim_id = node_dict[prim_id]
                else:
                    print(f'node {prim_id} not found in topology, skipping')
                    continue
                print(f'loading node {prim_id} from kb for {key}')
                response = self.session.post(f'{self._url}{prim_id}/load_kb',
                                                headers={'X-API-KEY': self._api_key},
                                                json={'data': value})
                if response.status_code != 200:
                    print(f'error loading kb for node {prim_id}: {response.text}')
                    return 'failed'
                print(f'loading node {prim_id} succeeded')

        return 'success'

    @_ensure_connected
    def put_kbs(self, archive_file): # pragma: no cover
        """Uploads KBs from local archive_file file.

        .. warning::

            .. deprecated:: 12.0 (Kehl)

            Use :func:`load_kbs_from_json` instead.

        """
        _headers = self._headers
        _headers['Content-Encoding'] = 'gzip'
        _headers['Content-type'] = 'application/octet-stream'
        with open(archive_file, 'rb') as f:
            data = f.read()
            response = self.session.put(f'{self._url}database',
                                        headers=_headers,
                                        verify=self._verify,
                                        data=data
                                        ).json()
        return response

    def set_target_class(self, target_class: str, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Provide a target_class symbol for the searcher to look for.

        The searcher will ignore all other classes. This is a symbol that
        is in the last event of a classification sequence.

        Example:
                .. code-block:: python

                    agent.set_target_class(target_class="apple", nodes=["P1"])

        .. note::

            This will result in the nodes specified only predicting on models
            that include the target class
        """
        if nodes is None:
            nodes = self.query_nodes
        return self._query(self.session.post,
                           'set-target-class',
                           nodes=nodes,
                           data=target_class)

    def clear_target_class(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Clears target selection.

            Example:
                .. code-block:: python

                    agent.clear_target_class(nodes=["P1"])
        """
        if nodes is None:
            nodes = self.query_nodes
        return self._query(self.session.post,
                           'clear-target-class',
                           nodes=nodes)

    @_ensure_connected
    def get_symbol(self, symbol_name: str, nodes: List = None) -> Union[dict, Any]:
        """Returns symbol with name *symbol_name*.

        Args:
            symbol_name (str): Name of symbol to retrieve
            nodes (List, optional): list of nodes to retrieve symbol information from.
                Defaults to None (all nodes).

        Returns:
            Union[dict, Any]: dictionary of {node_name : symbol} pairs
        Example:
                .. code-block:: python

                    agent.get_symbol("WEATHER|CLOUDY")

        """
        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.get(f"{self._url}{node['id']}/symbol/{symbol_name}",
                                            headers=self._headers,
                                            verify=self._verify
                                            ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return response['message']
                else:
                    result[node['name']] = response['message']
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None

        return result
    
    @_ensure_connected
    def clear_snapshots(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Clear the WM snapshots of the Agent.
        Will not delete any persisted memory from the KnowledgeBase

        Example:
            .. code-block:: python

                >>> agent.clear_snapshots(nodes=['P1'])
                'snapshots-cleared'
        """
        if nodes is None:
            nodes = self.all_nodes
        return self._query(self.session.post, 'snapshots/clear', nodes=nodes)

    @_ensure_connected
    def get_prediction_ensemble_for_model(self, model_name : str, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Get a prediction ensemble for the model with passed hash.

        Example:
            .. code-block:: python

                >>> agent.get_prediction_ensemble_for_model(model_name="a2s51cas6541321htkmy56", nodes=['P1'])
                'an ensemble or empty list'
        """
        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.get(f"{self._url}{node['id']}/prediction-ensemble-for-model/{model_name}",
                                            headers=self._headers,
                                            verify=self._verify
                                            ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return response['message']
                else:
                    result[node['name']] = response['message']
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None

        return result

    @_ensure_connected
    def get_ensembles_for_all_models(self, nodes: List = None) -> Union[dict, Tuple[dict, str]]:
        """Get ensembles for all models in Nodes

        Example:
            .. code-block:: python

                >>> agent.get_ensembles_for_all_models( nodes=['P1'])
                'a list of ensembles or empty dict'
        """
        if nodes is None:
            nodes = self.all_nodes
        else:
            nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
        result = {}
        for node in nodes:
            try:
                response = self.session.get(f"{self._url}{node['id']}/prediction-ensemble-for-model",
                                            headers=self._headers,
                                            verify=self._verify
                                            ).json()
                if 'error' in response or response['status'] == 'failed':
                    if len(nodes) == 1 and self.summarize_for_single_node:
                        raise AgentQueryError(response)
                    warnings.warn(response["message"], AgentQueryWarning)
                if len(nodes) == 1 and self.summarize_for_single_node:
                    return response['message']
                else:
                    result[node['name']] = response['message']
            except Exception as exception:
                raise AgentQueryError(str(exception)) from None

        return result

    @_ensure_connected
    def investigate_record(self, node, record: str) -> dict:
        """Function for drilling down on a model recursively in a multi-node
        agent. Will retrieve all data "under" the abstracted model,
        giving a complete representation of the abstracted model

        Args:
            node (list or str): the node to start at/containing
                the model provided in `record`: e.g. ['P4'] or 'P4'
            record (str): the model to search for

        Returns:
            dict: recursive structure containing a complete representation
            of the abstracted model
        """
        data = {}
        try:
            prev_summarize_state = self.summarize_for_single_node
            self.set_summarize_for_single_node(True)
            if isinstance(node, list):
                node = node[0]

            # if the node passed is a primitive_id, get the name
            # (P1, P2, etc.) of the node
            if node in self.genome.primitives:
                node = self.genome.primitives[node]['name']

            return self.__get_model_details(model=record,
                                            node=[node],
                                            topLevel=True)

        except Exception as e:
            print(str(e))
            traceback.print_exc()
        finally:
            self.set_summarize_for_single_node(prev_summarize_state)
        return data

    @_ensure_connected
    def investigate(self, record: str) -> dict:
        """Higher level implementation of :func:`investigate_record` that
        does not require a node to be entered. Operates on abstracted
        symbols, models, etc. Useful when applying in a programatic
        manner, where the node a symbol comes from is unknown.
        May throw exception if unable to identify a starting point.

        Args:
            record (str): the symbol to look for

        Returns:
            dict: recursive structure containing a complete representation
            of the abstracted model

        Example:
            .. code-block:: python

                record = 'PRIMITIVE|p124b23f9|004da01b0ef40d7c3e0965a6b4fd0f413672fbff|matches|KEY|VALUE'
                traceback = agent.investigate(record=record)
        """
        try:

            split_record = str(record).split('|')
            if split_record[0] == 'PRIMITIVE':
                if len(split_record) < 3:
                    raise Exception(f'Investigate does not know what to do with symbol: {record}')

                # print(f'investigating model {split_record[2]} on node {split_record[1]}')
                output = self.investigate_record(record=split_record[2],
                                                 node=split_record[1])
                if not output:
                    return {}
                query = {
                    'record': record,
                    'node_id': split_record[1],
                    'name': split_record[2]
                }
                return {'query': query,
                        'results': output
                        }
            else:
                prev_summarize_state = self.summarize_for_single_node
                try:
                    self.set_summarize_for_single_node(False)
                    tmp_model = self.get_model(record)
                    self.set_summarize_for_single_node(True)
                    actual_model_dict = {}
                    # print(f'{tmp_model = }')
                    for k, v in tmp_model.items():
                        if v:
                            actual_model_dict[k] = v
                    # print(f'{actual_model_dict = }')
                    if len(actual_model_dict) == 1:
                        for k, v in actual_model_dict.items():
                            # print(f'{k = }, {v = }')
                            output = self.investigate_record(record=v['name'],
                                                             node=[k])
                            query = {
                                'record': v['name'],
                                'node_id': k,
                                'name': v['name']
                            }

                            return {'query': query,
                                    'results': output}
                except Exception as e:
                    print(f'Exception in investigate call: {e}')
                finally:
                    self.set_summarize_for_single_node(prev_summarize_state)
        except Exception as e:
            print(f'Exception in investigate call: {e}')
        return {}

    @_ensure_connected
    def __get_model_details(self, model, node, topLevel=False):
        """Recursive function to get details about a specific
        model on a specific node

        Args:
            model (str): the model hash to look for
            node (list(str)): The node to look for the model on (provided as
                a list with a single element)
            topLevel (bool, optional): flag used to show that the function
                is at the top level (no recursing). Defaults to False.

        Returns:
            dict: recursive structure representing the data from the
            topLevel model
        """
        if model is None:
            return None
        p_id = self.genome.get_primitive_map().get(node[0], None)
        if p_id is None:
            return None

        p_name = p_id
        if p_id in self.genome.primitives:
            p_name = self.genome.primitives[p_id]['name']
        # print(f'{p_id=}, {p_name=}')
        if topLevel is True:
            # top level request
            record_data = {'record': f"{model}",
                           'subitems': [],
                           'model': self.get_model(model, nodes=node),
                           'node': p_name,
                           'node_id': p_id,
                           'topLevel': topLevel,
                           'bottomLevel': False}
        else:
            record_data = {'record': f"PRIMITIVE|{p_id}|{model['name']}",
                           'subitems': [],
                           'model': model,
                           'node': p_name,
                           'node_id': p_id,
                           'topLevel': topLevel,
                           'bottomLevel': False}

        for event_list in record_data['model']['sequence']:
            event = []
            for item in event_list:
                symbol = {}
                split_model = item.split('|')
                if split_model[0] == 'PRIMITIVE':
                    node_id = split_model[1]

                    node_name = node_id
                    if node_id in self.genome.primitives:
                        node_name = self.genome.primitives[node_id]['name']

                    sub_model = self.get_model(split_model[2],
                                               nodes=[node_name])
                    symbol = self.__get_model_details(model=sub_model,
                                                      node=[node_name])
                elif split_model[0] == 'VECTOR':
                    symbol = {'record': item,
                              'data': self.get_vector(item, nodes=node),
                              'subitems': None,
                              'bottomLevel': True,
                              'topLevel': False,
                              'node': node[0],
                              'node_id': p_id}
                else:
                    # print(f'reached symbol: {item}')
                    # if 'VECTOR' in item:
                    #     symbol = {'record': item,
                    #               'data': self.get_vector(item, nodes=node),
                    #               'subitems': None,
                    #               'bottomLevel': True,
                    #               'topLevel': False,
                    #               'node': node[0],
                    #               'node_id': p_id}
                    # else:
                    symbol = item
                event.append(deepcopy(symbol))
            record_data["subitems"].append(event)
        if tuple(record_data["subitems"]) == tuple(record_data["model"]["sequence"]):
            del record_data["subitems"]

        if 'subitems' in record_data:
            unique_subitems = []
            for subitem in record_data['subitems']:
                if subitem not in unique_subitems:
                    unique_subitems.append(subitem)

            record_data["subitems"] = tuple(unique_subitems)
        else:
            record_data['bottomLevel'] = True

        return record_data



#!/usr/bin/env python3
from optparse import OptionParser
import os
import sys
import json
import logging
import socket
import random
import docker
import pathlib
from typing import Union
from docker.models.networks import Network
from docker.models.containers import Container
import docker.errors as de
from sty import fg, bg
from retry import retry
import ia
from ia.gaius.genome_info import Genome

logging.basicConfig(format='%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
                    datefmt='%Y-%m-%d:%H:%M:%S')
logger = logging.getLogger(__name__)

ASSETS = {'network': {},
          'agent': None,
          'genome': None,
          'interfaces': {},
          'cp-containers': {},
          'mp-containers': {},
          # These are connected containers as applications. Ex. JIA notebooks.
          'applications': {}
          }


def retrieve_genome(genome_file: Union[str, dict]) -> Genome:
    logger.debug(f'{genome_file = }')
    
    logger.debug(f'{genome_file}, {type(genome_file)}')
    if isinstance(genome_file, dict):
        _g = genome_file

    # if provided filepath, attempt to open genome
    elif os.path.exists(pathlib.Path(genome_file)):
        with open(genome_file, 'r') as f:
            # read data into genome_json
            _g = json.load(f)

    # otherwise try to load genome from string
    else:
        try:
            _g = json.loads(genome_file)
        except json.decoder.JSONDecodeError:
            logger.critical(f'Failed to retreive genome, {genome_file=}')
            sys.exit(1)

    # remove description fields from genome
    for d in _g['elements']['nodes']:
        if 'description' in d['data'].keys():
            logger.debug(f"Removing: {d['data']['description']}")
            d['data']['description'] = ''

    return Genome(_g)


def parse_genome(genome):

    targets_map = {}  # values are target IDs for each key
    sources_map = {}  # values are source IDs of each key

    if 'edges' not in genome.topology['elements']:
        logger.debug("No connections.")
    else:
        for edge in genome.topology['elements']['edges']:
            d = edge['data']
            if d['source'] not in targets_map:
                if d['source'].startswith("action"):
                    continue
                targets_map[d['source']] = [d['target']]
            else:
                if d['source'].startswith("action"):
                    continue
                targets_map[d['source']].append(d['target'])
            if d['target'] not in sources_map:
                if d['target'].startswith("action"):
                    continue
                sources_map[d['target']] = [d['source']]
            else:
                if d['target'].startswith("action"):
                    continue
                sources_map[d['target']].append(d['source'])

    targets_map = eval(f"{targets_map}".replace(
        ", 'P1_ID': ['action1_ID', 'action2_ID']", ""))

    as_inputs = {pid: [] for pid in genome.primitives.keys()}
    is_input = []

    for pid, p in genome.primitives.items():
        logger.debug(f"{pid} ({p['name']}):")
        logger.debug(f"    {p['sources']}")
        if p['sources'] == ['observables']:
            as_inputs[pid] = []
        else:
            for mid in p['manipulatives']:
                if mid.startswith("action"):
                    continue
                logger.debug(
                    f"   {mid} {genome.manipulatives[mid]['name']} {genome.manipulatives[mid]['genes']['sources']['value']}")
                if genome.manipulatives[mid]['genes']['sources']['value'] == ['observables']:
                    as_inputs[pid].append(mid)
                    is_input.append(mid)

    logger.debug(f'{as_inputs=}')
    logger.debug(f'{is_input=}')

    logger.debug(f'{targets_map=}')
    # for any manipulative that is an observables (i.e. in is_input list),
    # remove that manipulative from targets of other manipulatives:
    for _id, targets in targets_map.items():
        for mid in is_input:
            logger.debug(f"Need to pop: {mid} from {targets}")
            while mid in targets:
                logger.debug(f"popping...{mid} from {_id}")
                targets_map[_id].pop(targets.index(mid))

    logger.debug("PROCESSING OBSERVABLES MANIPULATIVES")
    for mid, x in genome.manipulatives.items():
        if ['observables'] == x['genes']['sources']['value']:
            sources_map[mid] = [x['primitive']]

    logger.debug("sources_map:")
    for k, v in sources_map.items():
        logger.debug(f"{k}:")
        for t in v:
            if t.startswith('action'):
                v.pop(v.index(t))
            else:
                logger.debug(f"    <---{t}")

    logger.debug("targets_map:")
    for k, v in targets_map.items():
        logger.debug(f"{k}:")
        for t in v:
            if t.startswith('action'):
                v.pop(v.index(t))
            else:
                logger.debug(f"    --->{t}")

    logger.debug("as_inputs:")
    for k, v in as_inputs.items():
        logger.debug(f"{k}:")
        for t in v:
            if t.startswith('action'):
                v.pop(v.index(t))
            else:
                logger.debug(f"    --->({t})")

    return sources_map, targets_map, as_inputs


def check_container_conflicts(p_ids, running_container_names):
    docker_client = docker.from_env()
    logger.debug(f'checking container conflicts')
    logger.debug(f'{p_ids=}')
    for _id in p_ids:
        try:
            c = docker_client.containers.get(_id)
        except de.NotFound as e:
            continue
        x = input(
            f"{bg.red}{fg.white} Conflicting containers! {_id} already exists.{bg.rs}{fg.rs} Okay to replace? Y/N: ")
        if x.lower() == 'y':
            c = docker_client.containers.get(_id)
            c.stop()
            c.remove()
        else:
            logger.debug(f"{bg.yellow}{fg.white} Aborting!{fg.rs}{bg.rs}")
            sys.exit()
    return


def find_available_ports():
    """Search all docker containers, looking for free ports to assign
    to GAIuS API, starting at port 8000 (unsecured) and 44300 (secured)

    Returns:
        dict: showing ports found
    """
    docker_client = docker.from_env()
    port_connects = {}
    port_offset = 0
    port_base = 8000
    port_limit = 8200
    port_list = []

    for each in docker_client.containers.list():
        for ports in each.ports.values():
            if ports is None:
                continue
            for port in ports:
                port_list.append(int(port['HostPort']))
    while port_base + port_offset < port_limit:
        if (port_base + port_offset not in port_list) and (44300 + port_offset not in port_list):
            port_connects = {'80/tcp': port_base +
                             port_offset, '443/tcp': 44300+port_offset}
            break
        else:
            port_offset += 1

    if port_base + port_offset >= port_limit:
        logger.debug(f'failed to assign ports: reached port limit')
        sys.exit(1)
    return port_connects


def multi_container_start(genome: Genome,
                          BOTTLE_HOSTNAME=None,
                          REGISTRY=None,
                          VERSION=None,
                          API_KEY=None,
                          LOG_LEVEL=None,
                          NETWORK=None,
                          BIND_PORTS=None,
                          user_id=None,
                          agent_id=None):

    if user_id is None:
        raise Exception('no user id provided')
    if agent_id is None:
        raise Exception('no agent_id provided')

    docker_client = docker.from_env()
    container_extension = '-' + user_id + '-' + agent_id

    if not NETWORK:
        NETWORK = "g2network"

    net_name = NETWORK + container_extension
    # Does this network already exist?
    net_name_exists = False
    for n in docker_client.networks.list():
        if n.name == net_name:
            net_name_exists = True
            net = n
    if net_name_exists:
        logger.warning(f"error, {net_name} already exists!")
    else:
        net = docker_client.networks.create(net_name)
        logger.debug(
            f"{fg.green} Created {net.name} with id: {net.short_id} {fg.rs}")
    logger.debug(net_name)
    logger.debug(net)
    ASSETS['network'][net_name] = net

    sources_map, targets_map, as_inputs = parse_genome(genome)

    running_container_names = [c.name for c in docker_client.containers.list()]
    p_ids = list(genome.manipulative_map.keys()) + \
        list(genome.primitive_map.values()) + ['gaius-api']
    p_ids = [f'{_id}{container_extension}' for _id in p_ids]
    check_container_conflicts(p_ids, running_container_names)

    logger.debug("Starting Manipulative nodes...")
    for _id, m in genome.manipulative_map.items():
        if _id.startswith('action'):  # Needed for old testing topologies.
            continue
        manipulatives_genes = json.dumps(genome.manipulatives[_id])
        if _id in sources_map:
            sources = json.dumps(sources_map[_id])
        else:
            sources = json.dumps([])
        if _id in targets_map:
            targets = json.dumps(targets_map[_id])
        else:
            targets = json.dumps([])

        ASSETS['mp-containers'][_id] = docker_client.containers.run(
            name=_id + container_extension,
            image=f"{REGISTRY}manipulative-processor:{VERSION}",
            network=net_name,
            detach=True,
            init=True,
            hostname=_id,
            restart_policy={"Name": "unless-stopped"},
            environment={'LOG_LEVEL': LOG_LEVEL,
                         'API_KEY': API_KEY,
                         'HOSTNAME': _id,
                         'MANIFEST': manipulatives_genes,
                         'SOURCES': sources,
                         'TARGETS': targets}
        )

    logger.debug("Starting Cognitive Processor nodes...")
    for p, _id in genome.primitive_map.items():
        primitive_genes = json.dumps(genome.primitives[_id])
        if _id in sources_map:
            sources = json.dumps(sources_map[_id])
        else:
            sources = json.dumps([])
        if _id in targets_map:
            targets = json.dumps(targets_map[_id])
        else:
            targets = json.dumps([])
        if _id in as_inputs:
            inputs = json.dumps(as_inputs[_id])
        else:
            inputs = json.dumps([])

        image = f"{REGISTRY}cognitive-processor:{VERSION}"
        logger.debug(f"using cognitive processor image {image}")
        ASSETS['cp-containers'][_id] = docker_client.containers.run(
            name=_id + container_extension,
            image=image,
            network=net_name,
            detach=True,
            init=True,
            hostname=_id,
            restart_policy={"Name": "unless-stopped"},
            environment={'LOG_LEVEL': LOG_LEVEL,
                         'API_KEY': API_KEY,
                         'HOSTNAME': _id,
                         'MANIFEST': primitive_genes,
                         'SOURCES': sources,
                         'AS_INPUTS': inputs,
                         'TARGETS': targets}
        )

    logger.debug("Starting GAIuS-API interface node...")

    port_connects = spawn_gaius_api(docker_client=docker_client,
                                    container_extension=container_extension,
                                    REGISTRY=REGISTRY,
                                    VERSION=VERSION,
                                    net_name=net_name,
                                    BIND_PORTS=BIND_PORTS,
                                    LOG_LEVEL=LOG_LEVEL,
                                    API_KEY=API_KEY,
                                    BOTTLE_HOSTNAME=BOTTLE_HOSTNAME,
                                    genome=genome)
    # port_connects = {}
    # if BIND_PORTS:
    #     port_connects = find_available_ports()
    # ASSETS['interfaces']['gaius-api' + container_extension] = docker_client.containers.run(name='gaius-api' + container_extension,
    #                                                                                         image=f"{REGISTRY}gaius-api:{VERSION}",
    #                                                                                         network=net_name,
    #                                                                                         detach=True,
    #                                                                                         init=True,
    #                                                                                         ports=port_connects,
    #                                                                                         restart_policy={
    #                                                                                             "Name": "unless-stopped"},
    #                                                                                         environment={'LOG_LEVEL': LOG_LEVEL,
    #                                                                                                     'API_KEY': API_KEY,
    #                                                                                                     'HOSTNAME': BOTTLE_HOSTNAME,
    #                                                                                                     'GENOME': json.dumps(genome.topology),
    #                                                                                                     'PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION': "cpp"}
    #                                                                                         )

    # print summary
    logger.info(
        f'{fg.green}GAIuS Agent started on docker network {fg.blue}{str(net_name)}{fg.rs}')
    if (BIND_PORTS):
        logger.info(
            f'{fg.green} Ports {fg.blue}{str(port_connects)}{fg.green} are exposed to the host{fg.rs}')
    else:
        logger.warning(
            f'{fg.green}Ports were not exposed for this agent. re-run command with {fg.blue}--bind-ports{fg.green} to expose ports{fg.rs}')

    return net_name, port_connects


# retry spawning up to 5 times, with up to 1 second random delay between retries
# when multiple agents are being spawned simultaneously, the delay between finding available ports
# and starting the docker container is too large -> ports become no longer available
@retry(de.APIError, tries=5, delay=random.random())
def spawn_gaius_api(docker_client: docker.DockerClient,
                    container_extension: str,
                    REGISTRY: str,
                    VERSION,
                    net_name,
                    BIND_PORTS,
                    LOG_LEVEL,
                    API_KEY,
                    BOTTLE_HOSTNAME,
                    genome) -> dict:

    port_connects = {}
    if BIND_PORTS:
        port_connects = find_available_ports()
    try:
        ASSETS['interfaces']['gaius-api' + container_extension] = docker_client.containers.run(name='gaius-api' + container_extension,
                                                                                               image=f"{REGISTRY}gaius-api:{VERSION}",
                                                                                               network=net_name,
                                                                                               detach=True,
                                                                                               init=True,
                                                                                               ports=port_connects,
                                                                                               restart_policy={
                                                                                                   "Name": "unless-stopped"},
                                                                                               environment={'LOG_LEVEL': LOG_LEVEL,
                                                                                                            'API_KEY': API_KEY,
                                                                                                            'HOSTNAME': BOTTLE_HOSTNAME,
                                                                                                            'GENOME': json.dumps(genome.topology),
                                                                                                            'PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION': "cpp"}
                                                                                               )
    except de.APIError as spawn_exception:
        logger.error(
            f"error spawning gaius-api{container_extension}: {str(spawn_exception)}. Retrying")
        try:
            container = docker_client.containers.get(
                'gaius-api' + container_extension)
            container.kill()
        except de.NotFound:
            pass
        raise spawn_exception

    return port_connects


def single_container_start(genome: Genome,
                           BOTTLE_HOSTNAME=None,
                           REGISTRY=None,
                           VERSION=None,
                           API_KEY=None,
                           LOG_LEVEL=None,
                           NETWORK=None,
                           BIND_PORTS=None,
                           user_id=None,
                           agent_id=None):

    if user_id is None:
        raise Exception('no user id provided')
    if agent_id is None:
        raise Exception('no agent_id provided')

    docker_client = docker.from_env()
    container_extension = '-' + user_id + '-' + agent_id

    if not NETWORK:
        NETWORK = "g2network"

    net_name = NETWORK + container_extension
    # Does this network already exist?
    net_name_exists = False
    for n in docker_client.networks.list():
        if n.name == net_name:
            net_name_exists = True
            net = n
    if net_name_exists:
        logger.warning(f"error, {net_name} already exists!")
    else:
        net = docker_client.networks.create(net_name)
        logger.debug(
            f"{fg.green} Created {net.name} with id: {net.short_id} {fg.rs}")
    logger.debug(net_name)
    logger.debug(net)
    ASSETS['network'][net_name] = net

    sources_map, targets_map, as_inputs = parse_genome(genome)

    running_container_names = [genome.topology["agent"]]
    p_ids = list(genome.manipulative_map.keys()) + \
        list(genome.primitive_map.values()) + ['gaius-api']
    p_ids = [f'{_id}{container_extension}' for _id in p_ids]
    check_container_conflicts(p_ids, running_container_names)

    logger.debug("Starting Cognitive Processor nodes...")
    for p, _id in genome.primitive_map.items():
        primitive_genes = json.dumps(genome.primitives[_id])
        if _id in sources_map:
            sources = json.dumps(sources_map[_id])
        else:
            sources = json.dumps([])
        if _id in targets_map:
            targets = json.dumps(targets_map[_id])
        else:
            targets = json.dumps([])
        if _id in as_inputs:
            inputs = json.dumps(as_inputs[_id])
        else:
            inputs = json.dumps([])

    logger.debug("Starting GAIuS-API interface node...")
    port_connects = {}
    port_offset = 0
    port_base = 8000
    port_limit = 8200
    if (BIND_PORTS):

        while port_base + port_offset < port_limit:
            try:
                tmpSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                avail = tmpSocket.connect_ex(
                    ("127.0.0.1", port_base+port_offset))
                if (avail != 0):
                    avail = tmpSocket.connect_ex(
                        ("127.0.0.1", 44300+port_offset))
                    if (avail != 0):
                        logger.debug(
                            f"port {port_base+port_offset} not used. using {port_offset} as offset")
                        break
                port_offset = port_offset + 1
                tmpSocket.close()
                if (port_base + port_offset >= port_limit):
                    logger.debug(
                        f"ran out of available ports in range. first available port > {port_limit}")
                    sys.exit(-1)
            except ConnectionRefusedError:
                logger.debug("Connection refused when searching for ports :(")
                break
            finally:
                tmpSocket.close()

        port_connects = {'80/tcp': port_base +
                         port_offset, '443/tcp': 44300+port_offset}
        logger.debug(port_connects)

    logger.debug(f"{genome.primitive_map.keys()}")
    inputs = [val for key, val in genome.primitive_map.items()]

    '''
        This section aligns the genes and data according to name, and then
        preprocesses it so that it can be correctly read by the C++ code.
    '''
    # get primitive names
    primitive_names = str([val for key, val in genome.primitive_map.items()])
    primitive_names = primitive_names.replace("[", "")
    primitive_names = primitive_names.replace("]", "")
    primitive_names = primitive_names.replace("'", "")
    primitive_names = primitive_names.replace(" ", "")
    primitive_names = primitive_names.replace(",", ";")

    manip_names = str([key for key in genome.manipulative_map.keys()])
    manip_names = manip_names.replace("[", "")
    manip_names = manip_names.replace("]", "")
    manip_names = manip_names.replace("'", "")
    manip_names = manip_names.replace(" ", "")
    manip_names = manip_names.replace(",", ";")

    logger.debug(f"{primitive_names=}")
    logger.debug(f"{manip_names=}")

    primitive_genes = str([genome.primitives[key] for key in inputs])

    logger.debug(f"{inputs=}")

    primitive_genes = primitive_genes[1:-1]
    # primitive_genes = inputs.replace("'", "")
    primitive_genes = primitive_genes.replace(" ", "")
    primitive_genes = primitive_genes.replace("},", "};")
    primitive_genes = primitive_genes.replace("'", "\"")
    primitive_genes = primitive_genes.replace("False", "false")
    primitive_genes = primitive_genes.replace("True", "true")
    # logger.debug(f"{genome.primitive_map.items()=}")
    logger.debug(f"{primitive_genes=}")
    logger.debug(f"pre {targets_map=}")
    targets = str([targets_map[key] if key in targets_map else []
                  for key in inputs])
    targets = targets.replace("[[", "[")
    targets = targets.replace("]]", "]")
    targets = targets.replace(" ", "")
    targets = targets.replace("\'", "\"")
    targets = targets.replace(",", ";")
    logger.debug(f"post {targets=}")
    logger.debug(f"pre {sources_map=}")
    sources = str([sources_map[key] if key in sources_map else []
                  for key in inputs])
    sources = sources.replace("[[", "[")
    sources = sources.replace("]]", "]")
    sources = sources.replace(" ", "")
    sources = sources.replace("\'", "\"")
    sources = sources.replace("],", "];")
    logger.debug(f"post {sources=}")

    logger.debug(f"{as_inputs=}")

    inputs = str([[as_inputs[key]] if key in as_inputs else []
                 for key in inputs])
    inputs = inputs.replace(" ", "")
    inputs = inputs.replace("[[", "[")
    inputs = inputs.replace("]]", "]")
    inputs = inputs.replace("'", "\"")
    inputs = inputs.replace(",", ";")
    inputs = inputs.replace("[[", "[")
    inputs = inputs.replace("]]", "]")

    logger.debug(f"final {inputs=}")

    # create manipulative environment variables
    logger.debug(f"{genome.manipulatives.keys()=}")
    for val, key in genome.manipulative_map.items():
        logger.debug(f"{val=}, {key=}")
        genome.primitive_map[val] = val
        genome.primitives[val] = genome.manipulatives[val]

    manipulatives_genes = {}
    manip_ids = []
    manip_sources = {}
    manip_targets = {}
    for _id, m in genome.manipulative_map.items():
        logger.debug(f"{_id=}, {m=}")
        manip_ids.append(_id)
        if _id.startswith('action'):  # Needed for old testing topologies.
            continue
        manipulatives_genes[_id] = genome.manipulatives[_id]
        if _id in sources_map:
            manip_sources[_id] = sources_map[_id]
        else:
            manip_sources[_id] = []
        if _id in targets_map:
            manip_targets[_id] = targets_map[_id]
        else:
            manip_targets[_id] = []

    manipulatives_genes = [val for val in genome.manipulatives.values()]

    manipulatives_genes = str(manipulatives_genes)
    logger.debug(f"pre processing {manipulatives_genes=}")
    manipulatives_genes = manipulatives_genes[1:-1]
    # primitive_genes = inputs.replace("'", "")
    manipulatives_genes = manipulatives_genes.replace("}, {", "}; {")
    manipulatives_genes = manipulatives_genes.replace(" ", "")
    manipulatives_genes = manipulatives_genes.replace("'", "\"")
    # manipulatives_genes = manipulatives_genes.replace("False", "false")
    # manipulatives_genes = manipulatives_genes.replace("True", "true")
    manip_targets = str([manip_targets[key] for key in manip_ids])
    manip_targets = manip_targets.replace("[[", "[")
    manip_targets = manip_targets.replace("]]", "]")
    manip_targets = manip_targets.replace(" ", "")
    manip_targets = manip_targets.replace("\'", "\"")
    manip_targets = manip_targets.replace(",", ";")
    manip_sources = str([manip_sources[key] for key in manip_ids])
    manip_sources = manip_sources.replace("[[", "[")
    manip_sources = manip_sources.replace("]]", "]")
    manip_sources = manip_sources.replace(" ", "")
    manip_sources = manip_sources.replace("\'", "\"")
    manip_sources = manip_sources.replace("],", "];")
    manip_ids = str(manip_ids)
    manip_ids = manip_ids.replace("[", "")
    manip_ids = manip_ids.replace("]", "")
    manip_ids = manip_ids.replace("'", "")
    manip_ids = manip_ids.replace(" ", "")
    manip_ids = manip_ids.replace(",", ";")

    # logger.debug(f"{manipulatives_genes=}")
    logger.debug(f"{manip_ids=}")
    logger.debug(f"{manip_targets=}")
    logger.debug(f"{manip_sources=}")

    ASSETS['cp-containers'][_id] = docker_client.containers.run(name='gaius-api' + container_extension,
                                                                image=f"{REGISTRY}connectome:{VERSION}",
                                                                network=net_name,
                                                                detach=True,
                                                                init=True,
                                                                hostname=BOTTLE_HOSTNAME,
                                                                ports=port_connects,
                                                                ipc_mode="shareable",
                                                                privileged=True,
                                                                volumes=[
                                                                    "/dev/bus/usb/:/dev/bus/usb/"],
                                                                # publish_all_ports=True,
                                                                restart_policy={
                                                                    "Name": "unless-stopped"},
                                                                environment={'LOG_LEVEL': LOG_LEVEL,
                                                                             'API_KEY': API_KEY,
                                                                             'HOSTNAME': BOTTLE_HOSTNAME,
                                                                             'CONTAINER_NAME': BOTTLE_HOSTNAME,
                                                                             'PRIM_NAMES': primitive_names,
                                                                             'PRIM_MANIFEST': primitive_genes,
                                                                             'PRIM_SOURCES': sources,
                                                                             'PRIM_AS_INPUTS': inputs,
                                                                             'PRIM_TARGETS': targets,
                                                                             'GENOME': json.dumps(genome.topology),
                                                                             'MANIPNAMES': manip_names,
                                                                             'MANIPIDS': manip_ids,
                                                                             'MANIPMANIFEST': manipulatives_genes,
                                                                             'MANIPSOURCES': manip_sources,
                                                                             'MANIPTARGETS': manip_targets}
                                                                )

    # print summary
    logger.info(
        f'{fg.green}GAIuS Agent started on docker network {fg.blue}{str(net_name)}{fg.rs}')
    if (BIND_PORTS):
        logger.info(
            f'{fg.green} Ports {fg.blue}{str(port_connects)}{fg.green} are exposed to the host{fg.rs}')
    else:
        logger.warning(
            f'{fg.green}Ports were not exposed for this agent. re-run command with {fg.blue}--bind-ports{fg.green} to expose ports{fg.rs}')

    return net_name, port_connects


def startit(genome_file,
            BOTTLE_HOSTNAME=None,
            REGISTRY=None,
            VERSION=None,
            API_KEY=None,
            LOG_LEVEL=None,
            NETWORK=None,
            BIND_PORTS=None,
            user_id=None,
            agent_id=None):

    genome = retrieve_genome(genome_file=genome_file)

    # logger.debug(f"{genome.topology.keys()=}")

    if (genome.topology.get("communication_protocol", "tcp") != "uds"):
        net_name, port_connects = multi_container_start(
            genome, BOTTLE_HOSTNAME, REGISTRY, VERSION, API_KEY, LOG_LEVEL, NETWORK, BIND_PORTS, user_id, agent_id)
    else:
        net_name, port_connects = single_container_start(
            genome, BOTTLE_HOSTNAME, REGISTRY, VERSION, API_KEY, LOG_LEVEL, NETWORK, BIND_PORTS, user_id, agent_id)

    return net_name, port_connects


def stopContainer(id: str, docker_client=None):
    try:
        if docker_client is None:
            docker_client = docker.from_env()
        logger.debug(f'''stopping container {id}''')
        c = docker_client.containers.get(id)
        c.stop()
        c.remove()
        logger.debug(f'''container {id} removed{fg.rs}''')
    except:
        logger.debug(
            f'''{fg.red}container {id} did not exist... continuing{fg.rs}''')


def stopit(genome_file, user_id: str = None, agent_id: str = None, NETWORK: str = None):

    if user_id is None:
        raise Exception('no user id provided')
    if agent_id is None:
        raise Exception('no agent_id provided')

    container_extension = '-' + user_id + '-' + agent_id
    docker_client = docker.from_env()

    genome = retrieve_genome(genome_file)

    ids = list(genome.primitive_map.values()) + \
        list(genome.manipulative_map.keys())
    ids.append('gaius-api')

    ids = ['{0}{1}'.format(id, container_extension) for id in ids]
    ids.append(genome.topology["agent"])  # for when single container is run

    for id in ids:
        stopContainer(id, docker_client=docker_client)

    container_extension = '-' + user_id + '-' + agent_id
    if not NETWORK:
        NETWORK = "g2network"
    net_name = NETWORK + container_extension

    delete_docker_network(docker_client=docker_client, network_name=net_name)

    try:
        logger.debug("pruning docker system")
        docker_client.containers.prune()
        docker_client.networks.prune()
    except:
        logger.warning("unable to prune docker")
        pass

    return


def delete_docker_network(docker_client: docker.DockerClient, network_name: str):
    """Delete single docker network, forcefully removing containers from network

    Args:
        docker_client (docker.DockerClient): Docker Client
        network_name (str): Name of network to be deleted
    """
    logger.debug(f'Attempting to forcefully delete docker network {network_name}')
    try:
        network: Network = docker_client.networks.get(network_name)
        network.reload()
        container_list = network.containers

        # remove all remaining containers from docker network (forcefully)
        container: Container
        for container in container_list:
            try:
                network.disconnect(container=container, force=True)
            except de.APIError:
                pass # if container already disconnected
        network.remove()
    except de.NotFound:
        logger.debug("Network %s was not found", network_name)
        pass

    except de.APIError:
        logger.exception(
            "API Error when deleting docker network %s", network_name)

    return


def execute_agent(BOTTLE_HOSTNAME=None,
                  REGISTRY='registry.digitalocean.com/intelligent-artifacts',
                  VERSION='latest',
                  API_KEY=None,
                  LOG_LEVEL='DEBUG',
                  NETWORK=None,
                  BIND_PORTS=True,
                  GENOME=None,
                  AGENT_ID=None,
                  USER_ID=None,
                  KILL=False
                  ):

    if GENOME is None:
        raise Exception('No genome provided')

    # if not os.path.exists(f"{os.path.expanduser('~')}/.gaius"):
    #     os.mkdir(f"{os.path.expanduser('~')}/.gaius")

    if not API_KEY:
        API_KEY = 'ABCD-1234'

    if not NETWORK:
        NETWORK = 'g2network'

    configuration = {
        'BOTTLE_HOSTNAME': BOTTLE_HOSTNAME,
        'REGISTRY': REGISTRY,
        'VERSION': VERSION,
        'API_KEY': API_KEY,
        'LOG_LEVEL': LOG_LEVEL,
        'NETWORK': NETWORK,
        'BIND_PORTS': BIND_PORTS,
        'user_id': USER_ID,
        'agent_id': AGENT_ID,
    }

    genome_file = GENOME

    if KILL:
        stopit(genome_file, user_id=USER_ID,
               agent_id=AGENT_ID, NETWORK=NETWORK)
        return -1

    _, ports = startit(genome_file, **configuration)

    return ports


def start_agent(json_obj: dict):
    '''Start an agent from a genome passed as JSON

    fields => example:

        genome => json representation of genome
        user_id => test_user
        agent_id => agent1
        kill => False
        version => "latest"

    '''

    # required fields
    genome = json_obj['genome']
    user_id = json_obj['user_id']
    agent_id = json_obj['agent_id']
    kill = json_obj['kill']
    network = json_obj['network']

    # optional fields
    api_key = json_obj.get('api_key', "ABCD-1234")
    registry = json_obj.get(
        'registry', 'registry.digitalocean.com/intelligent-artifacts/')
    version = json_obj.get('version', 'latest')

    agent_name = 'gaius-api' + "-" + user_id + "-" + agent_id

    ports = execute_agent(REGISTRY=registry,
                          BIND_PORTS=True,
                          AGENT_ID=agent_id,
                          USER_ID=user_id,
                          KILL=kill,
                          API_KEY=api_key,
                          GENOME=json.dumps(genome),
                          VERSION=version,
                          NETWORK=network)

    if kill:
        return -1

    logger.debug(f'started agent with ports {ports}')

    agent_info = {"name": "",
                  "domain": agent_name,
                  "api_key": api_key,
                  "secure": False}

    return agent_info, ports


def options():
    global options
    parser = OptionParser(version=ia.__version__)
    parser.add_option("-g", "--genome", dest="filename",
                      help="Genome file location", metavar="FILE")

    parser.add_option("--api-key", dest="API_KEY", default='ABCD-1234',
                      help="""Secret API key.
                        Default = 'ABCD-1234'
                                """)

    parser.add_option("--log-level", dest="LOG_LEVEL", default='DEBUG',
                      help="""Log level.
                        Default = DEBUG
                        Options:
                            CRITICAL
                            ERROR
                            WARNING
                            INFO
                            DEBUG
                                """)

    parser.add_option("-n", "--network", dest="NETWORK",
                      help="""Network in which all nodes will communicate.
                        Default = 'g2network'
                                """)
    parser.add_option("--bind-ports", action="store_true", dest="BIND_PORTS", default=False,
                      help="""To enable multiple agents running on single system, port are not bound by default.
                        Specify this option to explicitly bind agent to ports (required to run tests on agent).
                            HTTP port: 8000 + agent_id
                            HTTPS port: 44300 + agent_id
                            """)

    parser.add_option("-b", "--gaius-api-name", dest="BOTTLE", default='gaius-api',
                      help="""Bottle name
                        Default = 'BOTTLE'
                                """)

    parser.add_option("-H", "--hostname", dest="BOTTLE_HOSTNAME", default='localhost',
                      help="""Bottle name
                        Default = 'BOTTLE'
                                """)

    parser.add_option("-r", "--registry", dest="REGISTRY", default='registry.digitalocean.com/intelligent-artifacts/',
                      help="""Container registry
                        Default = 'registry.digitalocean.com/intelligent-artifacts/'

                        Set to '' if you want to use only local containers.
                                """)

    parser.add_option("--container-version", dest="VERSION", default='latest',
                      help="""Version of container images. Same used for all.
                        Default = 'latest'
                        """)

    parser.add_option("-d", "--debug",
                      action="store_true", dest="debug", default=False,
                      help="print debug information")

    parser.add_option("--kill",
                      action="store_true", dest="kill", default=False,
                      help="Kill all agent's containers.")

    parser.add_option("--agent-id",
                      dest="agent_id", default="1",
                      help="""A string that uniquely identifies the agent and network. Appended to container and network names
                             Default=1""")

    parser.add_option("--user-id",
                      dest="user_id", default=os.environ.get("USER"),
                      help="""A string that uniquely identifies the agent and network. Appended to container and network names
                             Default=os.environ.get("USER")""")

    (options, args) = parser.parse_args()

    if not options.filename:
        logger.warn(
            "Need a genome file location! Pass with -g or --genome arg.")
        sys.exit(1)

    if len(options.REGISTRY) > 0 and not options.REGISTRY.endswith('/'):
        options.REGISTRY += '/'


if __name__ == '__main__':
    # if not os.path.exists(f"{os.path.expanduser('~')}/.gaius"):
    #     os.mkdir(f"{os.path.expanduser('~')}/.gaius")
    options()

    DEBUG = options.debug
    if DEBUG:
        logger.setLevel(level=logging.DEBUG)

    # Default values provided by OptionParser
    BOTTLE_HOSTNAME = options.BOTTLE_HOSTNAME
    REGISTRY = options.REGISTRY
    VERSION = options.VERSION
    API_KEY = options.API_KEY
    LOG_LEVEL = options.LOG_LEVEL
    if (options.NETWORK):
        NETWORK = options.NETWORK
    else:
        NETWORK = "g2network"

    BIND_PORTS = options.BIND_PORTS

    USER_ID = options.user_id
    AGENT_ID = options.agent_id
    configuration = {
        'BOTTLE_HOSTNAME': BOTTLE_HOSTNAME,
        'REGISTRY': REGISTRY,
        'VERSION': VERSION,
        'API_KEY': API_KEY,
        'LOG_LEVEL': LOG_LEVEL,
        'NETWORK': NETWORK,
        'BIND_PORTS': BIND_PORTS,
        'user_id': USER_ID,
        'agent_id': AGENT_ID,
    }

    genome_file = options.filename
    if options.kill:
        stopit(genome_file, user_id=USER_ID, agent_id=AGENT_ID)
        sys.exit(0)

    network = startit(genome_file, **configuration)
